import importlib
import inspect
from abc import ABC, abstractmethod
from typing import Dict, Any, Type, Tuple, List

import somax.features
from somax import features
from somax.utils.introspective import StringParsed, Introspective


class AbstractFeature(Introspective, ABC):
    def __init__(self, value: Any):
        self._value = value

    @abstractmethod
    def value(self) -> Any:
        """ TODO: docstring """

    @classmethod
    def classes(cls, include_abstract: bool = False) -> List[Type['AbstractFeature']]:
        return list(cls._classes(somax.features, include_abstract=include_abstract).values())

    def name(self) -> str:
        return self.__class__.__name__


class CorpusFeature(AbstractFeature, ABC):

    @classmethod
    @abstractmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        pass

    # @classmethod
    # @abstractmethod
    # def analyze(cls, event: 'CorpusEvent', fg_spectrogram: Spectrogram, bg_spectrogram: Spectrogram,
    #             fg_chromagram: Chromagram, bg_chromagram: Chromagram, **kwargs):
    #     # TODO: Add previous events as optional argument too
    #     pass

    @classmethod
    @abstractmethod
    def decode(cls, trait_dict: Dict[str, Any]) -> 'CorpusFeature':
        """ TODO docstring """

    @abstractmethod
    def encode(self) -> Dict[str, Any]:
        """ TODO docstring """

    @classmethod
    def classpath(cls) -> str:
        return cls.__module__ + "." + cls.__name__

    @staticmethod
    def from_json(classpath: str, feature_kwargs: Dict[str, Any]) -> Tuple[Type['CorpusFeature'], 'CorpusFeature']:
        """ Raises: KeyError, AttributeError"""
        module_name, class_name = classpath.rsplit(".", 1)
        cls: Type[CorpusFeature] = getattr(importlib.import_module(module_name), class_name)
        return cls, cls.decode(feature_kwargs)

    @staticmethod
    def all_corpus_features() -> List[Tuple[str, Type['CorpusFeature']]]:
        return inspect.getmembers(features, lambda m: inspect.isclass(m)
                                                      and not inspect.isabstract(m)
                                                      and issubclass(m, CorpusFeature))


class RuntimeFeature(AbstractFeature, StringParsed, ABC):
    @classmethod
    def from_string(cls, keyword: str, value: Any = None, **kwargs) -> 'RuntimeFeature':
        """ :raises ValueError if a feature matching the keyword doesn't exist """
        for feature in cls._classes(somax.features).values():  # type: Type[RuntimeFeature]
            if issubclass(feature, RuntimeFeature) and feature.keyword() == keyword:
                return feature(value)
        raise ValueError(f"No feature matches the keyword '{keyword}'")

    @classmethod
    def default(cls, **kwargs) -> 'StringParsed':
        """ :raises ValueError. As no default Feature exists, this will always be raised"""
        raise ValueError("No default feature exist")

    @staticmethod
    @abstractmethod
    def keyword() -> str:
        """ The keyword that the class responds to when receiving runtime influences from the client. """



import logging
from abc import abstractmethod, ABC
from typing import List, Optional

import numpy as np
from scipy import sparse

from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.improvisation_memory import ImprovisationMemory
from somax.runtime.parameter import Parametric, Parameter
from somax.runtime.peaks import Peaks
from somax.runtime.transforms import AbstractTransform
from somax.utils.introspective import StringParsed


class AbstractMergeAction(Parametric, StringParsed, ABC):

    def __init__(self):
        super().__init__()

    @abstractmethod
    def merge(self, peaks: Peaks, time: float, history: ImprovisationMemory = None, corpus: Corpus = None,
              **kwargs) -> Peaks:
        """ """

    @abstractmethod
    def feedback(self, feedback_event: CorpusEvent, time: float, applied_transform: AbstractTransform) -> None:
        """ """

    @abstractmethod
    def clear(self) -> None:
        """ """

    @classmethod
    def default(cls, **_kwargs) -> 'AbstractMergeAction':
        return DistanceMergeAction()

    @classmethod
    def from_string(cls, merge_action: str, **kwargs) -> 'AbstractMergeAction':
        return cls._from_string(merge_action, **kwargs)


class DistanceMergeAction(AbstractMergeAction):

    def __init__(self, t_width: float = 0.09):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.logger.debug(f"[__init__] Creating DistanceMergeAction with width {t_width}.")
        self._t_width: Parameter = Parameter(t_width, 0.0, None, 'float', "Very unclear parameter")  # TODO
        self._parse_parameters()

    def __repr__(self):
        return f"{type(self).__name__}(_t_width={self.t_width})"

    def merge(self, peaks: Peaks, _time: float, _history: ImprovisationMemory = None, corpus: Corpus = None,
              **_kwargs) -> Peaks:
        if peaks.size() <= 1:
            return peaks
        self.logger.debug(f"[merge] Merging activity with {peaks.size()} peaks.")

        duration: float = corpus.duration()
        inv_duration: float = 1 / duration
        num_rows: int = int(duration / self._t_width.value)

        peaks_list: List[Peaks] = []
        for transform_hash in np.unique(peaks.transform_ids):
            indices: np.ndarray = np.argwhere(peaks.transform_ids == transform_hash)
            indices = indices.reshape((indices.size,))
            scores: np.ndarray = peaks.scores[indices]
            times: np.ndarray = peaks.times[indices]
            num_cols: int = scores.size

            row_indices: np.ndarray = np.floor(times * inv_duration * num_rows).astype(np.int32)
            interp_matrix: sparse.coo_matrix = sparse.coo_matrix(
                (np.ones(num_cols), (row_indices, np.arange(num_cols))),
                shape=(num_rows + 1, num_cols))
            interp_matrix: sparse.csc_matrix = interp_matrix.tocsc()

            interpolated_scores: np.ndarray = interp_matrix.dot(scores)
            interpolated_times: np.ndarray = interp_matrix.dot(times)
            num_peaks_per_index: np.ndarray = np.array(interp_matrix.sum(axis=1)).reshape(interp_matrix.shape[0])
            peak_indices: np.ndarray = interpolated_scores.nonzero()[0]

            scores: np.ndarray = interpolated_scores[peak_indices]
            times: np.ndarray = np.divide(interpolated_times[peak_indices], num_peaks_per_index[peak_indices])
            transforms: np.ndarray = np.ones(peak_indices.size, dtype=np.int32) * transform_hash
            # print("After merge:", scores.shape, times.shape, transforms.shape)

            peaks_list.append(Peaks(scores, times, transforms))

        merged_peaks: Peaks = Peaks.concatenate(peaks_list)
        self.logger.debug(f"[merge] Merge successful. Number of peaks after merge: {merged_peaks.size()}.")
        return merged_peaks

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float, applied_transform: AbstractTransform) -> None:
        pass

    def clear(self) -> None:
        pass

    @property
    def t_width(self):
        return self._t_width.value

    @t_width.setter
    def t_width(self, value):
        self._t_width.value = value





import logging
from typing import Optional, Any, List

from somax.runtime.corpus import ContentType
from somax.runtime.corpus_event import CorpusEvent, Note
from somax.runtime.exceptions import InvalidCorpus
from somax.runtime.player import Player
from somax.runtime.transforms import AbstractTransform
from somax.scheduler.base_scheduler import BaseScheduler
from somax.scheduler.scheduled_event import ScheduledEvent, AutomaticTriggerEvent, ManualTriggerEvent, TempoEvent, \
    AbstractTriggerEvent, MidiEvent, AudioEvent, ScheduledInfluenceEvent, ScheduledCorpusEvent, NewStateEvent
from somax.scheduler.scheduled_object import TriggerMode


class AgentScheduler(BaseScheduler):
    TRIGGER_PRETIME: float = 0.01  # seconds

    def __init__(self, player: Player, tempo: float = 120.0, trigger_pretime: float = TRIGGER_PRETIME,
                 tempo_master: bool = False, initial_tick: float = 0.0, running: bool = False):
        super().__init__(tempo=tempo, tick=initial_tick, running=running)
        self.logger = logging.getLogger(__name__)
        self._player = player
        self.queue: List[ScheduledEvent] = []
        self.is_tempo_master: bool = tempo_master
        self._trigger_pretime: float = trigger_pretime

    ######################################################
    # MAIN FUNCTIONS
    ######################################################

    def update_tick(self, tick: float, tempo: float) -> List[ScheduledEvent]:
        if self.running:
            self._update_tick(tick=tick, tempo=tempo)
            events: List[ScheduledEvent] = self._process_internal_events()
            return events
        else:
            return []

    def _update_tick(self, tick: float, tempo: float, **_kwargs):
        self._tick = tick
        self._tempo = tempo

    ######################################################
    # ADD (INTERNAL AND EXTERNAL)
    ######################################################

    def add_trigger_event(self, trigger_time: Optional[float] = None):
        if self._player.trigger_mode == TriggerMode.AUTOMATIC and not self._has_trigger():
            self._add_automatic_trigger_event(self._tick - self._trigger_pretime * self.tempo / 60.0, self._tick)
        elif self._player.trigger_mode == TriggerMode.MANUAL and self.running:
            self._add_manual_trigger_event(trigger_time if trigger_time else self._tick)
        else:
            self.logger.debug("[add_trigger_event] Could not add trigger.")

    def _add_automatic_trigger_event(self, trigger_time: float, target_time: float):
        self.queue.append(AutomaticTriggerEvent(trigger_time, target_time))

    def _add_manual_trigger_event(self, trigger_time: float):
        self.queue.append(ManualTriggerEvent(trigger_time))

    def add_tempo_event(self, trigger_time: float, tempo: float):
        self.queue.append(TempoEvent(trigger_time, tempo))

    def _add_corpus_event(self, trigger_time: float, corpus_event: CorpusEvent, applied_transform: AbstractTransform):
        if self.is_tempo_master:
            self.add_tempo_event(trigger_time, corpus_event.tempo)

        if self._player.corpus.content_type == ContentType.AUDIO:
            self._add_audio_event(trigger_time, corpus_event)
        elif self._player.corpus.content_type == ContentType.MIDI:
            self._add_midi_event(trigger_time, corpus_event, applied_transform)

        self._add_new_state_event(trigger_time, corpus_event, applied_transform)

    def _add_midi_event(self, trigger_time: float, corpus_event: CorpusEvent, applied_transform: AbstractTransform):
        # Handle held notes from previous state:
        note_offs_previous: List[Note] = [n for n in self._player.held_notes if n not in corpus_event.held_to()]
        note_ons: List[Note] = [n for n in corpus_event.notes if n not in self._player.held_notes]
        note_offs: List[Note] = [n for n in corpus_event.notes if n not in corpus_event.held_from()]
        self._player.held_notes = corpus_event.held_from()

        # Queue midi events for note ons/offs
        for note in self._player.artificially_held_notes:
            onset: float = trigger_time
            self.queue.append(MidiEvent(onset, note.pitch, 0, note.channel,
                                        corpus_event.state_index, applied_transform))
        for note in note_offs_previous:
            onset: float = trigger_time
            self.queue.append(MidiEvent(onset, note.pitch, 0, note.channel, corpus_event.state_index, None))
        for note in note_ons:
            if self._player.simultaneous_onsets:
                onset: float = trigger_time
            else:
                onset: float = trigger_time + max(0.0, note.onset)
            self.queue.append(
                MidiEvent(onset, note.pitch, note.velocity, note.channel, corpus_event.state_index, applied_transform))
        if self._player.hold_notes_artificially:
            self._player.artificially_held_notes = note_offs
        else:
            for note in note_offs:
                onset: float = trigger_time + max(0.0, note.onset + note.duration)
                self.queue.append(MidiEvent(onset, note.pitch, 0, note.channel, corpus_event.state_index, None))

    def _add_audio_event(self, trigger_time: float, corpus_event: CorpusEvent):
        raise NotImplementedError

    def _add_new_state_event(self, trigger_time: float, corpus_event, applied_transform: AbstractTransform):
        self.queue.append(NewStateEvent(trigger_time, corpus_event, applied_transform))

    ######################################################
    # PROCESS (INTERNAL)
    ######################################################

    def _process_internal_events(self) -> List[ScheduledEvent]:
        events: List[ScheduledEvent] = [e for e in self.queue if e.trigger_time <= self._tick]
        self.queue = [e for e in self.queue if e.trigger_time > self._tick]
        # sort to ensure that influences/midi/tempo/etc. are handled before triggers if simultaneous
        events.sort(key=lambda e: isinstance(e, AbstractTriggerEvent))
        target_events: List[ScheduledEvent] = []
        for event in events:
            if isinstance(event, TempoEvent) and self.is_tempo_master:
                target_events.append(event)
            elif isinstance(event, MidiEvent):
                target_events.append(event)
            elif isinstance(event, AudioEvent):
                target_events.append(event)
            elif isinstance(event, NewStateEvent):
                target_events.append(event)
            elif isinstance(event, AbstractTriggerEvent):
                self._process_trigger_event(event)
            elif isinstance(event, ScheduledInfluenceEvent):
                self._process_influence_event(event)
            elif isinstance(event, ScheduledCorpusEvent):
                self._process_corpus_event(event)
        return target_events

    def _process_trigger_event(self, trigger_event: AbstractTriggerEvent) -> None:
        try:
            event_and_transform: Optional[tuple[CorpusEvent, AbstractTransform]]
            # By default, target time should always be the time given by the trigger_event, but may occasionally use
            #   self.tick to handle unexpected delays in scheduling (for example while loading a corpus)
            trigger_event.target_time = max(trigger_event.target_time, self.tick)
            trigger_event.trigger_time = max(trigger_event.trigger_time, self.tick - self._trigger_pretime)
            event_and_transform = self._player.new_event(trigger_event.target_time, self._tempo)
        except InvalidCorpus as e:
            self.logger.error(str(e))
            self._requeue_trigger_event(trigger_event)
            return

        if event_and_transform is None:
            if self._player.trigger_mode == TriggerMode.AUTOMATIC:
                self._requeue_trigger_event(trigger_event)
            return

        event: CorpusEvent = event_and_transform[0]
        applied_transform: AbstractTransform = event_and_transform[1]
        self._add_corpus_event(trigger_event.target_time, event, applied_transform)

        if isinstance(trigger_event, AutomaticTriggerEvent) and self._player.trigger_mode == TriggerMode.AUTOMATIC:
            if event.duration > 0:
                next_trigger_time: float = trigger_event.trigger_time + event.duration
                next_target_time: float = trigger_event.target_time + event.duration
                self._add_automatic_trigger_event(next_trigger_time, next_target_time)
            else:
                self._requeue_trigger_event(trigger_event)

    def _process_corpus_event(self, corpus_event: ScheduledCorpusEvent) -> Optional[Any]:
        raise AttributeError(f"Queued Corpus Events are not supported for class {self.__class__.__name__}.")

    def _process_influence_event(self, influence_event: ScheduledInfluenceEvent):
        raise AttributeError(f"Queued Influence Events are not supported for class {self.__class__.__name__}.")

    ######################################################
    # OTHER
    ######################################################

    def delete_trigger(self):
        self.queue = [e for e in self.queue if not (isinstance(e, AutomaticTriggerEvent))]

    def _requeue_trigger_event(self, trigger_event: AbstractTriggerEvent) -> None:
        next_trigger_time: float = trigger_event.trigger_time + 1
        next_target_time: float = trigger_event.target_time + 1
        self._add_automatic_trigger_event(next_trigger_time, next_target_time)

    def _has_trigger(self) -> bool:  # TODO: Unoptimized approach
        for event in self.queue:
            if isinstance(event, AutomaticTriggerEvent):
                return True
        return False

    def _sanity_check(self):
        pass  # TODO

    ######################################################
    # REAL-TIME CONTROL
    ######################################################

    def start(self, **kwargs):
        self.running = True

    def pause(self, **kwargs):
        self.running = False

    def stop(self, **kwargs) -> List[ScheduledEvent]:
        self.running = False
        self._tick = 0
        return self.flush()

    def flush(self) -> List[ScheduledEvent]:
        remamining_queue: List[ScheduledEvent] = self.queue[:]
        self.queue = []
        target_events: List[ScheduledEvent] = []
        for event in remamining_queue[:]:
            # Add new triggers for all existing automatically triggered
            if isinstance(event, AutomaticTriggerEvent):
                self.add_trigger_event()
            # output queued note offs
            if isinstance(event, MidiEvent) and event.velocity == 0:
                target_events.append(event)

        # handle held notes
        for note in self._player.held_notes:
            target_events.append(MidiEvent(self._tick, note.pitch, 0, note.channel))
        for note in self._player.artificially_held_notes:
            target_events.append(MidiEvent(self._tick, note.pitch, 0, note.channel))

        self._player.held_notes = []
        self._player.artificially_held_notes = []
        return target_events



import logging
from abc import abstractmethod, ABC


class BaseScheduler(ABC):
    def __init__(self, tempo: float = 120.0, tick: float = 0.0, running: bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)
        self.tempo: float = tempo
        self._tick: float = tick
        self.running: bool = running

    @abstractmethod
    def start(self, **kwargs):
        pass

    @abstractmethod
    def pause(self, **kwargs):
        pass

    @abstractmethod
    def stop(self, **kwargs):
        pass

    @property
    def tick(self):
        return self._tick


# class AsyncScheduler(BaseScheduler, ABC):
#     DEFAULT_CALLBACK_INTERVAL: float = 0.001  # seconds
#
#     def __init__(self, tempo: float = 120.0, *args, **kwargs):
#         super().__init__(tempo=tempo, *args, **kwargs)
#         self._last_callback_time: float = time.time()
#         self.terminated: bool = False
#
#     async def init_async_loop(self, callback_interval: int = DEFAULT_CALLBACK_INTERVAL):
#         self.logger.debug(f"Scheduler initialized with callback interval {callback_interval}.")
#         while not self.terminated:
#             await asyncio.sleep(callback_interval)
#             self._callback()
#
#     @abstractmethod
#     def _callback(self):
#         pass

    # def _update_tick(self, **_kwargs):
    #     if self.running:
    #         t: float = time.time()
    #         delta_time: float = t - self._last_callback_time
    #         self._last_callback_time = t
    #         self._tick += delta_time * self.tempo / 60.0



import inspect
import logging
import sys
from abc import abstractmethod, ABC
from collections import deque
from typing import Tuple, Dict, Union, Optional, List, Type

from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import TransformError
from somax.runtime.label import AbstractLabel
from somax.runtime.parameter import Parameter, ParamWithSetter
from somax.runtime.parameter import Parametric
from somax.runtime.peak_event import PeakEvent, ClassicPeakEvent
from somax.runtime.transforms import AbstractTransform
from somax.runtime.transform_handler import TransformHandler
from somax.utils.introspective import StringParsed


class AbstractMemorySpace(Parametric, StringParsed, ABC):
    """ MemorySpaces determine how events are matched to labels """

    def __init__(self, **_kwargs):
        """ Note: kwargs can be used if additional information is need to construct the data structure.
            Note: labels are not classified in default constructor as additional parameters might need init before."""
        super(AbstractMemorySpace, self).__init__()
        self.logger = logging.getLogger(__name__)
        self._corpus: Optional[Corpus] = None
        self._labels: Optional[List[AbstractLabel]] = None
        # Must be defined in `update_transforms`
        self._transform_handler: Optional[TransformHandler] = None

    @classmethod
    def default(cls, **_kwargs) -> 'AbstractMemorySpace':
        return NGramMemorySpace()

    @classmethod
    def from_string(cls, memory_space: str, **kwargs) -> 'AbstractMemorySpace':
        return cls._from_string(memory_space, **kwargs)

    @abstractmethod
    def model(self, corpus: Corpus, labels: List[AbstractLabel], **_kwargs) -> None:
        pass

    @abstractmethod
    def influence(self, labels: List[Tuple[AbstractLabel, AbstractTransform]], time: float,
                  **kwargs) -> List[PeakEvent]:
        pass

    @staticmethod
    def classes() -> Dict[str, Type]:
        """Returns class objects for all non-abstract classes in this module."""
        return dict(inspect.getmembers(sys.modules[__name__],
                                       lambda member: inspect.isclass(member) and
                                                      not inspect.isabstract(member) and member.__module__ == __name__))

    @abstractmethod
    def clear(self) -> None:
        """ Reset the playing state of the memory space without removing its corpus memory. """

    @abstractmethod
    def update_transforms(self, transform_handler: TransformHandler, valid_transforms: List[AbstractTransform]):
        """ Note: Must define `self._transform_handler` as a reference to an object owned by a `Player` instance"""

    def update_parameter_dict(self) -> Dict[str, Union[Parametric, Parameter, Dict]]:
        parameters: Dict = {}
        for name, parameter in self._parse_parameters().items():
            parameters[name] = parameter.update_parameter_dict()
        self.parameter_dict = {"parameters": parameters}
        return self.parameter_dict

    def _remodel(self) -> None:
        if self._corpus and self._labels:
            self.model(self._corpus, self._labels)


class NGramMemorySpace(AbstractMemorySpace):
    def __init__(self, history_len: int = 3, **_kwargs):
        super(NGramMemorySpace, self).__init__(**_kwargs)
        self.logger.debug(f"[__init__] Initializing {self.__class__.__name__} with history length {history_len}.")
        self._structured_data: Dict[Tuple[int, ...], List[CorpusEvent]] = {}
        self._ngram_size: Parameter = ParamWithSetter(history_len, 1, None, 'int',
                                                      "Number of events to hard-match. (TODO)",
                                                      self.set_ngram_size)  # TODO
        # history per transform stored with transform id as key
        self._influence_history: Dict[int, deque[int]] = {}
        self._parse_parameters()

    def __repr__(self):
        return f"NGramMemorySpace(_ngram_size={self._ngram_size.value}, corpus={self._corpus}, ...)"  #

    def model(self, corpus: Corpus, labels: List[AbstractLabel], **_kwargs) -> None:
        if self._transform_handler is None:
            raise TransformError("update_transforms must be called before modelling a corpus")
        self.logger.debug(f"[model] Modelling corpus '{corpus.name}'.")
        self._corpus = corpus
        self._labels = labels
        self._structured_data = {}
        labels: deque[int] = deque([], self._ngram_size.value)
        for event, label in zip(self._corpus.events, self._labels):
            labels.append(hash(label))
            if len(labels) < self._ngram_size.value:
                continue
            else:
                key: Tuple[int, ...] = tuple(labels)
                value: CorpusEvent = event
                if key in self._structured_data:
                    self._structured_data[key].append(value)
                else:
                    self._structured_data[key] = [value]

    def influence(self, labels: List[Tuple[AbstractLabel, AbstractTransform]], time: float,
                  **_kwargs) -> List[PeakEvent]:
        matches: List[PeakEvent] = []
        for (label, transform) in labels:
            label_value: int = hash(label)
            transform_id: int = self._transform_handler.get_id(transform)
            self.logger.debug(f"[influence] Influencing memory space {self} with label {label_value}.")
            self._influence_history[transform_id].append(label_value)
            if len(self._influence_history[transform_id]) < self._ngram_size.value:
                # all deques will have the same length, hence if one is shorter than ngram size, all of them will be.
                return []
            else:
                key: Tuple[int, ...] = tuple(self._influence_history[transform_id])
                try:
                    matching_events: List[CorpusEvent] = self._structured_data[key]
                    for event in matching_events:
                        matches.append(ClassicPeakEvent(event, transform_id))
                except KeyError:  # no matches found
                    continue
        return matches

    def set_ngram_size(self, new_size: int):
        self._ngram_size.value = new_size
        self._influence_history = {t: deque([], new_size) for t in self._influence_history.keys()}
        self._remodel()

    def update_transforms(self, transform_handler: TransformHandler, valid_transforms: List[AbstractTransform]):
        if not valid_transforms:
            raise TransformError(f"No valid transforms exist in {self.__class__.__name__}")
        self._transform_handler = transform_handler
        transform_ids: List[int] = [transform_handler.get_id(t) for t in valid_transforms]
        self._influence_history = {t: deque([], self._ngram_size.value) for t in transform_ids}

    def clear(self) -> None:
        self._influence_history = {t: deque([], self._ngram_size.value) for t in self._influence_history.keys()}



import copy
import json
import logging
import os
from enum import Enum
from typing import List, Optional, Type, Dict, Any, Tuple, Union

from somax.settings import IMPORT_MATPLOTLIB

if IMPORT_MATPLOTLIB:
    pass
import numpy as np
import pandas as pd

from somax import settings
import somax
from somax.corpus_builder.chromagram import Chromagram
from somax.corpus_builder.matrix_keys import MatrixKeys as Keys
from somax.corpus_builder.note_matrix import NoteMatrix
from somax.corpus_builder.spectrogram import Spectrogram
from somax.features.feature import CorpusFeature
from somax.runtime.corpus_event import CorpusEvent, Note
from somax.runtime.exceptions import InvalidCorpus


class ContentType(Enum):
    MIDI = "MIDI"
    AUDIO = "Audio"

    def encode(self):
        return str(self)


class HeldObject:
    def __init__(self, note: Note, event: CorpusEvent):
        self.note = note
        self.event = event

    def __eq__(self, other):
        if isinstance(other, Note):
            return self.note == other
        if isinstance(other, HeldObject):
            return self.note == other.note
        raise AttributeError(f"Cannot compare type '{type(other)}' with '{self.__class__}'")


class Corpus:
    INDEX_MAP_SIZE = 1_000_000

    def __init__(self, events: List[CorpusEvent], name: str, content_type: ContentType,
                 build_parameters: Dict[str, Any], fg_spectrogram: Optional[Spectrogram] = None,
                 bg_spectrogram: Optional[Spectrogram] = None, fg_chromagram: Optional[Chromagram] = None,
                 bg_chromagram: Optional[Chromagram] = None):
        self.logger = logging.getLogger(__name__)
        self.events: List[CorpusEvent] = events
        # self.onsets: np.ndarray = np.array([e.onset for e in self.events])    # TODO: Remove (?)
        self.name: str = name
        self.content_type: ContentType = content_type
        self._build_parameters: Dict[str, Any] = build_parameters
        self._index_map: np.ndarray
        self._grid_size: int
        self._index_map, self._grid_size = Corpus._create_index_map(self.events, self.duration())

        # These parameters will not be stored when exported and will thus not exist in json-parsed corpora
        self.fg_spectrogram: Optional[Spectrogram] = fg_spectrogram
        self.bg_spectrogram: Optional[Spectrogram] = bg_spectrogram
        self.fg_chromagram: Optional[Chromagram] = fg_chromagram
        self.bg_chromagram: Optional[Chromagram] = bg_chromagram

    def __repr__(self):
        return f"Corpus(name='{self.name}', ...)"

    @classmethod
    def from_json(cls, filepath: str, volatile: bool = False) -> 'Corpus':
        """ Raises: IOError, InvalidCorpus"""
        try:
            with open(filepath, 'r') as f:
                corpus_data: Dict[str, Any] = json.load(f)
            version: str = corpus_data["version"]
            if version != somax.__version__ and not volatile:
                raise InvalidCorpus(f"The loaded corpus was built with an old version of Somax. "
                                    f"While it may work, using it could result in a number of bugs."
                                    f" Recommended action: rebuild corpus."
                                    f" (To attempt to load the corpus anyway: enable the 'volatile' flag)")
            name: str = corpus_data["name"]
            content_type: ContentType = ContentType(corpus_data["content_type"])

            build_parameters: Dict[str, Any] = corpus_data["build_parameters"]
            features_dict: Dict[str, str] = corpus_data["features_dict"]
            events: List[CorpusEvent] = [CorpusEvent.decode(event_dict, features_dict)
                                         for event_dict in corpus_data["events"]]
            return cls(events, name, content_type, build_parameters)

        # KeyError (from AbstractTrait.from_json, this), AttributeError (from AbstractTrait.from_json)
        except (KeyError, AttributeError) as e:
            raise InvalidCorpus(f"The Corpus at '{filepath}' has an invalid format and could not be loaded") from e

    def export(self, output_folder: str, overwrite: bool = False,
               indentation: Optional[int] = None) -> str:
        """ Raises IOError"""
        filepath = os.path.join(output_folder, self.name + ".json")
        if os.path.exists(filepath) and not overwrite:
            raise IOError(f"Could not export corpus as file '{filepath}' already exists. "
                          f"Set overwrite flag to True to overwrite existing.")
        else:
            with open(filepath, 'w') as f:
                json.dump(self, f, indent=indentation, default=lambda o: o.encode())
        return filepath

    def encode(self) -> Dict[str, Any]:
        features: Dict[Type['CorpusFeature'], str] = {cls: name for (name, cls) in CorpusFeature.all_corpus_features()}
        if len(set(features.keys())) < len(features.keys()):
            self.logger.warning("Two features with the same name exist in the library. Built corpus may not be properly"
                                " constructed. Ensure that no two CorpusFeatures in the library have the same name.")
        return {"name": self.name,
                "content_type": self.content_type.value,
                "version": somax.__version__,
                "build_parameters": self._build_parameters,
                "features_dict": {name: feature.classpath() for (feature, name) in features.items()},
                "length": self.length(),
                "duration": self.duration(),
                "events": [event.encode(features_dict=features) for event in self.events]
                }

    # def analyze(self, event_parameter: Type[CorpusFeature], **kwargs):
    #
    #     for event in self.events:
    #         parameter: CorpusFeature = event_parameter.analyze(event, self.fg_spectrogram, self.bg_spectrogram,
    #                                                            self.fg_chromagram, self.bg_chromagram, **kwargs)
    #         event.set_feature(parameter)

    def length(self) -> int:
        return len(self.events)

    def duration(self) -> float:
        if len(self.events) == 0:
            return 0.0
        last_event: CorpusEvent = self.events[-1]
        return last_event.onset + last_event.duration

    def event_at(self, index: int) -> CorpusEvent:
        return self.events[index]

    def event_around(self, time: float) -> CorpusEvent:
        index: int = self._index_map[int(np.floor(time * self._grid_size))]
        return self.event_at(index)

    def events_around(self, times: np.ndarray) -> List[CorpusEvent]:
        indices: np.ndarray = self._index_map[(np.floor(times * self._grid_size)).astype(int)]
        events: List[CorpusEvent] = [self.event_at(index) for index in indices]
        return events

    def to_note_matrix(self) -> NoteMatrix:
        note_data: List[List[Union[int, float, str]]] = [[] for _ in range(len(Keys))]
        held_previous: List[HeldObject] = []
        for event in self.events:
            held_current: List[HeldObject] = []
            for note in copy.deepcopy(event.notes):
                # Case 1: Note is held from a previous slice...
                if note in held_previous and note.is_held_to():
                    prev_held_note: HeldObject = [h for h in held_previous if h.note == note][0]
                    held_previous: List[HeldObject] = [h for h in held_previous if h.note != note]  # remove match
                    # Case 1.1: ...and held throughout the entire slice into the next one:
                    if note.is_held_from(event.duration):
                        prev_held_note.note.duration += event.duration  # Extend by entire slice duration
                        held_current.append(prev_held_note)
                    # Case 1.2: ...and ends in the current slice
                    else:
                        prev_held_note.note.duration += self._adjust_in_time(note, event).duration
                        note_data = self._append_to_matrix(prev_held_note.note, prev_held_note.event, note_data)
                # Case 2: Note starts in current slice but is held throughout the slice
                elif note.is_held_from(event.duration):
                    held_note: HeldObject = HeldObject(self._adjust_in_time(note, event), event)
                    held_current.append(held_note)
                # Case 3: Note starts (or is held_to but not held from previous slice) and ends within current slice
                else:
                    note_data = self._append_to_matrix(self._adjust_in_time(note, event), event, note_data)
            # Append all notes that were held from a previous slice but not extended into this one
            for held_note in held_previous:
                note_data = self._append_to_matrix(held_note.note, held_note.event, note_data)
            held_previous = held_current
        # Finally append any note held from the last slice
        for held_note in held_previous:
            note_data = self._append_to_matrix(held_note.note, held_note.event, note_data)

        note_matrix: pd.DataFrame = pd.DataFrame(np.array(note_data, dtype=object).T, columns=[key for key in Keys])
        note_matrix.sort_values([Keys.REL_ONSET, Keys.PITCH, Keys.CHANNEL], inplace=True)
        # note_matrix.drop_duplicates(subset=[Keys.PITCH, Keys.CHANNEL, Keys.REL_ONSET, Keys.REL_DURATION], inplace=True)
        return NoteMatrix(note_matrix)

    @staticmethod
    def _append_to_matrix(note: Note, event: CorpusEvent,
                          matrix: List[List[Union[int, float, str]]]) -> List[List[Union[int, float, str]]]:
        matrix[Keys.PITCH.value].append(note.pitch)
        matrix[Keys.VELOCITY.value].append(note.velocity)
        matrix[Keys.CHANNEL.value].append(note.channel)
        matrix[Keys.REL_ONSET.value].append(note.onset)
        matrix[Keys.REL_DURATION.value].append(note.duration)
        matrix[Keys.ABS_ONSET.value].append(note.absolute_onset)
        matrix[Keys.ABS_DURATION.value].append(note.absolute_duration)
        matrix[Keys.TEMPO.value].append(event.recorded_memory_state.tempo
                                        if event.recorded_memory_state is not None
                                        else event.tempo)
        matrix[Keys.BAR_NUMBER.value].append(event.bar_number)
        matrix[Keys.TRACK_NAME.value].append(note.track)
        return matrix

    @staticmethod
    def _adjust_in_time(note: Note, event: CorpusEvent):
        positive_part_note_onset: float = max(0.0, note.onset)
        positive_part_note_abs_onset: float = max(0.0, note.absolute_onset)
        if event.recorded_memory_state is not None:
            if event.recorded_memory_state.artificially_sustained:
                note.duration = event.duration - positive_part_note_onset
                note.absolute_duration = event.absolute_duration - positive_part_note_abs_onset
            if event.recorded_memory_state.simultaneous_onsets:
                note.onset = 0.0
                note.duration += positive_part_note_onset
                note.absolute_onset = 0.0
                note.absolute_duration += positive_part_note_abs_onset
        # transform note onset from relative (to slice) to absolute time since start of corpus (in ticks)
        note.onset = event.onset + positive_part_note_onset
        note.duration = min(note.duration, event.duration)
        note.absolute_onset = event.absolute_onset + positive_part_note_onset
        note.absolute_duration = min(note.absolute_duration, event.absolute_duration)
        return note

    # TODO: Removed for PyInstaller to exclude matplotlib
    # def plot(self):
    #     raise NotImplementedError("Plotting a corpus is currently not supported")
    #     import matplotlib as mpl
    #     mpl.use('Qt5Agg')
    #     if all([v is not None for v in
    #             [self.fg_spectrogram, self.bg_spectrogram, self.fg_chromagram, self.bg_chromagram]]):
    #         _, axes = plt.subplots(6, 1, gridspec_kw={'height_ratios': [1, 5, 3, 3, 1, 1]})
    #         self.fg_spectrogram.plot(axes[2])
    #         self.bg_spectrogram.plot(axes[3])
    #         self.fg_chromagram.plot(axes[4])
    #         self.bg_chromagram.plot(axes[5])
    #         self.to_note_matrix().plot(axes=(axes[0], axes[1]))
    #     else:
    #         _, axes = plt.subplots(2, 1, gridspec_kw={'height_ratios': [1, 5]})
    #         self.to_note_matrix().plot(axes=(axes[0], axes[1]))
    #     plt.show()

    @staticmethod
    def _create_index_map(events: List[CorpusEvent], corpus_duration_ticks: float) -> Tuple[np.ndarray, float]:
        grid_size: float = (Corpus.INDEX_MAP_SIZE - 1) / corpus_duration_ticks
        index_map: np.ndarray = np.zeros(Corpus.INDEX_MAP_SIZE, dtype=int)
        for event in events:
            start_index: int = int(np.floor(event.onset * grid_size))
            end_index: int = int(np.floor((event.onset + event.duration) * grid_size))
            index_map[start_index:end_index] = event.state_index
        return index_map, grid_size



from abc import ABC, abstractmethod
from typing import List, Tuple

import somax.classification
from somax.runtime.corpus import Corpus
from somax.runtime.influence import AbstractInfluence
from somax.runtime.label import AbstractLabel
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform
from somax.utils.introspective import StringParsed


class AbstractClassifier(StringParsed, ABC):

    def __init__(self, **kwargs):
        self._transforms: List[AbstractTransform] = []

    @classmethod
    def default(cls, **_kwargs) -> None:
        raise ValueError(f"No default classifier exists.")

    @classmethod
    def from_string(cls, name: str, **kwargs) -> 'AbstractClassifier':
        """ :raises ValueError"""
        return cls._from_string(name, module=somax.classification, **kwargs)

    @abstractmethod
    def cluster(self, corpus: Corpus) -> None:
        """ :raises InvalidCorpus if clustering fails."""
        pass

    @abstractmethod
    def classify_corpus(self, corpus: Corpus) -> List[AbstractLabel]:
        """ :returns List of untransformed labels of the same length as the corpus. """
        pass

    @abstractmethod
    def classify_influence(self, influence: AbstractInfluence) -> List[Tuple[AbstractLabel, AbstractTransform]]:
        """ :returns List of inverse transformed labels of the same length as number of `AbstractTransforms`
                     applied in atom. """
        pass

    @abstractmethod
    def clear(self) -> None:
        """ This function should reset any runtime state related to previous influences inside the classifier without
            unloading any data. If the classifier is stateless, leave this method blank."""
        pass

    @abstractmethod
    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :returns list of for the classifier valid `AbstractTransform`s selected from `transform_handler`. """

    # TODO: REMOVE
    # @classmethod
    # @abstractmethod
    # def rms(cls, influence_corpus: Corpus, output_corpus: Corpus) -> np.ndarray:
    #     """ The rms method is not strictly necessary to implement - it's only used for offline performance analysis.
    #         To leave it unimplemented, simply define it as
    #         ```
    #             @staticmethod
    #             def rms(influence_corpus: Corpus, output_corpus: Corpus) -> np.ndarray:
    #                 raise NotImplementedError(f"{self.__class__}.rms is not implemented")
    #         ```
    #     """
    #     pass



import logging
from logging import Handler

from somax.runtime.target import Target


class OscLogForwarder(Handler):

    def __init__(self, target: Target, logging_level: int = logging.INFO):
        super().__init__()
        self.target = target
        self.setLevel(logging_level)

    def emit(self, record):
        self.target.send(record.levelname.lower(), [self.format(record)])



import logging
from typing import List, Dict, Optional

from somax.classification.classifier import AbstractClassifier
from somax.runtime.activity_pattern import AbstractActivityPattern
from somax.runtime.atom import Atom
from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import DuplicateKeyError
from somax.runtime.improvisation_memory import ImprovisationMemory
from somax.runtime.memory_spaces import AbstractMemorySpace
from somax.runtime.merge_actions import AbstractMergeAction
from somax.runtime.parameter import Parameter
from somax.runtime.parameter import Parametric
from somax.runtime.peaks import Peaks
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform


class Streamview(Parametric):
    DEFAULT_WEIGHT = 1.0

    def __init__(self, name: str, transform_handler: TransformHandler, weight: float = DEFAULT_WEIGHT,
                 corpus: Optional[Corpus] = None, merge_action: AbstractMergeAction = AbstractMergeAction.default(),
                 **kwargs):
        super().__init__(**kwargs)
        self.logger = logging.getLogger(__name__)
        self.name: str = name
        self._transform_handler: TransformHandler = transform_handler  # passed as reference
        self.merge_action: AbstractMergeAction = merge_action

        self.atoms: Dict[str, Atom] = {}
        self.streamviews: Dict[str, Streamview] = {}

        self._weight: Parameter = Parameter(weight, 0.0, None, 'float', "Relative scaling of streamview peaks.")
        self.enabled: Parameter = Parameter(True, False, True, "bool", "Enables this Streamview.")
        self._parse_parameters()

        if corpus is not None:
            self.read_corpus(corpus)

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name}, ...)"

    ######################################################
    # MAIN RUNTIME FUNCTIONS
    ######################################################

    def _update_peaks_on_new_event(self, time: float) -> None:
        for streamview in self.streamviews.values():
            streamview._update_peaks_on_new_event(time)
        for atom in self.atoms.values():
            atom.update_peaks_on_new_event(time)

    def _merged_peaks(self, time: float, history: ImprovisationMemory, corpus: Corpus, **kwargs) -> Peaks:
        weight_sum: float = 0.0
        for streamview in self.streamviews.values():
            weight_sum += streamview.weight if streamview.is_enabled() else 0.0
        for atom in self.atoms.values():
            weight_sum += atom.weight if atom.is_enabled() else 0.0
        if weight_sum < 1e-6:
            self.logger.warning(f"Weights are invalid. Setting weight to 1.0 in streamview '{self.name}'.")
            weight_sum = 1.0

        peaks_list: List[Peaks] = []
        for streamview in self.streamviews.values():
            if streamview.is_enabled():
                peaks: Peaks = streamview._merged_peaks(time, history, corpus, **kwargs)
                peaks.scale(streamview.weight / weight_sum)
                peaks_list.append(peaks)
        for atom in self.atoms.values():
            if atom.is_enabled():
                peaks: Peaks = atom.get_peaks()
                peaks.scale(atom.weight / weight_sum)
                peaks_list.append(peaks)

        all_peaks: Peaks = Peaks.concatenate(peaks_list)
        return self.merge_action.merge(all_peaks, time, history, corpus, **kwargs)

    ######################################################
    # MODIFY STATE
    ######################################################

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        self.merge_action.feedback(feedback_event, time, applied_transform)
        for streamview in self.streamviews.values():
            streamview.feedback(feedback_event, time, applied_transform)
        for atom in self.atoms.values():
            atom.feedback(feedback_event, time, applied_transform)

    def clear(self) -> None:
        self.merge_action.clear()
        for streamview in self.streamviews.values():
            streamview.clear()
        for atom in self.atoms.values():
            atom.clear()

    def create_streamview(self, path: List[str], weight: float, merge_action: AbstractMergeAction,
                          override: bool = False) -> None:
        """ Raises KeyError, IndexError, DuplicateKeyError """
        new_streamview_name: str = path.pop(-1)  # raises: IndexError
        parent: Streamview = self._get_streamview(path)  # raises: KeyError, IndexError
        if new_streamview_name in parent.streamviews.keys() and not override:
            raise DuplicateKeyError(f"A Streamview with name '{new_streamview_name}' already exists."
                                    f"To override: use 'override=True'.")
        else:

            parent.streamviews[new_streamview_name] = Streamview(name=new_streamview_name, weight=weight,
                                                                 corpus=self.corpus, merge_action=merge_action)
            self._parse_parameters()

    def create_atom(self, path: List[str], weight: float, self_influenced: bool, classifier: AbstractClassifier,
                    activity_pattern: AbstractActivityPattern, memory_space: AbstractMemorySpace,
                    transforms: None = None, enabled: bool = True, override: bool = False) -> None:
        """ Raises KeyError, IndexError, DuplicateKeyError """
        new_atom_name: str = path.pop(-1)  # raise: IndexError
        parent_streamview: Streamview = self._get_streamview(path)  # raises: KeyError, IndexError
        if new_atom_name in parent_streamview.atoms.keys() and not override:
            raise DuplicateKeyError(f"An Atom with name '{new_atom_name}' already exists."
                                    f"To override: use 'override=True'.")
        else:
            self.atoms[new_atom_name] = Atom(name=new_atom_name, weight=weight, classifier=classifier,
                                             activity_pattern=activity_pattern, memory_space=memory_space,
                                             corpus=self.corpus, self_influenced=self_influenced, enabled=enabled)
            self._parse_parameters()

    def delete_atom(self, path: List[str]) -> None:
        """ Raises: KeyError, IndexError """
        atom_name: str = path.pop(-1)  # raises: IndexError
        parent_streamview: Streamview = self._get_streamview(path)  # raises: KeyError, IndexError
        del parent_streamview.atoms[atom_name]  # raises KeyError
        self._parse_parameters()

    def delete_streamview(self, path: List[str]) -> None:
        """ Raises: KeyError, IndexError """
        streamview_name: str = path.pop(-1)  # raises: IndexError
        parent_streamview: Streamview = self._get_streamview(path)  # raises: KeyError, IndexError
        del parent_streamview.streamviews[streamview_name]  # raises KeyError
        self._parse_parameters()

    def set_merge_action(self, path: List[str], merge_action: AbstractMergeAction) -> None:
        """ Raises: KeyError, IndexError """
        streamview: Streamview = self._get_streamview(path)  # raises: KeyError, IndexError
        streamview.merge_action = merge_action
        self._parse_parameters()

    def set_classifier(self, path: List[str], classifier: AbstractClassifier) -> None:
        """ Raises: KeyError, IndexError """
        atom: Atom = self._get_atom(path)  # raises: KeyError, IndexError
        atom.set_classifier(classifier)
        atom.update_transforms(self._transform_handler)
        self._parse_parameters()

    def set_memory_space(self, path: List[str], memory_space: AbstractMemorySpace) -> None:
        """ Raises: KeyError, IndexError """
        atom: Atom = self._get_atom(path)  # raises: KeyError, IndexError
        atom.set_memory_space(memory_space)
        atom.update_transforms(self._transform_handler)
        self._parse_parameters()

    def set_activity_pattern(self, path: List[str], activity_pattern: AbstractActivityPattern) -> None:
        """ Raises: KeyError, IndexError """
        atom: Atom = self._get_atom(path)  # raises: KeyError, IndexError
        atom.set_activity_pattern(activity_pattern)
        atom.update_transforms(self._transform_handler)
        self._parse_parameters()

    def read_corpus(self, corpus: Corpus) -> None:
        for streamview in self.streamviews.values():
            streamview.read_corpus(corpus)
        for atom in self.atoms.values():
            atom.read_corpus(corpus)

    def update_transforms(self):
        for streamview in self.streamviews.values():
            streamview.update_transforms()
        for atom in self.atoms.values():
            atom.update_transforms(self._transform_handler)

    def is_enabled(self) -> bool:
        return self.enabled.value

    @property
    def weight(self):
        return self._weight.value

    @weight.setter
    def weight(self, value: float):
        self._weight.value = value

    ######################################################
    # PRIVATE
    ######################################################

    def _get_streamview(self, path: List[str]) -> 'Streamview':
        """ Raises: KeyError, IndexError"""
        if not path:
            return self
        target_name: str = path.pop(0)
        if path:  # Path is not empty: descend recursively
            return self.streamviews[target_name]._get_streamview(path)
        else:
            return self.streamviews[target_name]

    def _get_atom(self, path: List[str]) -> Atom:
        """ Raises: KeyError, IndexError"""
        target_name: str = path.pop(0)
        if path:  # Path is not empty: descend recursively
            return self.streamviews[target_name]._get_atom(path)
        else:
            return self.atoms[target_name]

    def _direct_influenced_atoms(self) -> List[Atom]:
        return [atom for atom in self.all_atoms() if not atom.self_influenced]

    def _self_influenced_atoms(self) -> List[Atom]:
        return [atom for atom in self.all_atoms() if atom.self_influenced]

    def all_atoms(self) -> List[Atom]:
        return self.__gather_atoms([])

    def __gather_atoms(self, atoms: List[Atom]) -> List[Atom]:
        for streamview in self.streamviews.values():
            atoms = streamview.__gather_atoms(atoms)
        atoms.extend(self.atoms.values())
        return atoms



from abc import ABC, abstractmethod
from typing import List, Tuple

from somax.classification.classifier import AbstractClassifier
from somax.features import TopNote
from somax.features.feature import AbstractFeature
from somax.features.pitch_features import RuntimeIntegerPitch, AbstractIntegerPitch, VirtualFundamental, BassNote
from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import InvalidLabelInput, TransformError
from somax.runtime.influence import AbstractInfluence, CorpusInfluence, FeatureInfluence
from somax.runtime.label import IntLabel, AbstractLabel
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform, NoTransform


class BasicPitchClassifier(AbstractClassifier, ABC):
    def __init__(self):
        super().__init__()

    def cluster(self, corpus: Corpus, **kwargs) -> None:
        # No clustering required for class
        pass

    def classify_corpus(self, corpus: Corpus) -> List[IntLabel]:
        labels: List[IntLabel] = []
        for event in corpus.events:  # type: CorpusEvent
            labels.append(self._label_from_corpus_event(event, NoTransform()))
        return labels

    def classify_influence(self, influence: AbstractInfluence) -> List[Tuple[AbstractLabel, AbstractTransform]]:
        if isinstance(influence, FeatureInfluence) and isinstance(influence.feature, RuntimeIntegerPitch):
            return [(self._label_from_feature(influence.feature, t), t) for t in self._transforms]
        elif isinstance(influence, CorpusInfluence):
            # TODO: Handle or comment on KeyError, which technically should never occur
            return [(self._label_from_corpus_event(influence.corpus_event, t), t) for t in self._transforms]
        else:
            raise InvalidLabelInput(f"Influence {influence} could not be classified by {self}.")

    def clear(self) -> None:
        pass  # All basic pitch classifiers are stateless

    @staticmethod
    @abstractmethod
    def _label_from_corpus_event(event: CorpusEvent, transform: AbstractTransform) -> IntLabel:
        pass

    @staticmethod
    @abstractmethod
    def _label_from_feature(pitch: AbstractFeature, transform: AbstractTransform) -> IntLabel:
        pass

    @staticmethod
    @abstractmethod
    def _trait_from_corpus_event(event: CorpusEvent) -> int:
        pass


class TopNoteClassifier(BasicPitchClassifier):
    """ Classifies an event to its corresponding midi number in range [0, 127].
            Corpus:    Uses EventParameter TopNote in range [0, 127]
            Influence: Responds to keyword "pitch" in range [0, 127]. """

    @staticmethod
    def _label_from_corpus_event(event: CorpusEvent, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(event.get_feature(TopNote)).value()
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _label_from_feature(pitch: AbstractFeature, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(pitch).value()
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _trait_from_corpus_event(event: CorpusEvent) -> int:
        return event.get_feature(TopNote).value()

    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :raises TransformError if transform_handler doesn't contain any applicable transforms """
        self._transforms = transform_handler.get_by_feature(AbstractIntegerPitch)
        if not self._transforms:
            raise TransformError(f"No applicable transform exists in classifier {self.__class__}.")
        return self._transforms


class PitchClassClassifier(BasicPitchClassifier):
    # TODO: Clean up (lots of) code duplication
    """ Classifies an event to its corresponding pitch class (midi number % 12) in range [0, 11].
            Corpus: Uses EventParameter TopNote in range [0, 127]
            Influence: Responds to keyword "pitch" in range [0, 127]. """

    @staticmethod
    def _label_from_corpus_event(event: CorpusEvent, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(event.get_feature(TopNote)).value() % 12
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _label_from_feature(pitch: AbstractFeature, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(pitch).value() % 12
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _trait_from_corpus_event(event: CorpusEvent) -> int:
        return event.get_feature(TopNote).value() % 12

    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :raises TransformError if transform_handler doesn't contain any applicable transforms """
        self._transforms = transform_handler.get_by_feature(AbstractIntegerPitch)
        if not self._transforms:
            raise TransformError(f"No applicable transform exists in classifier {self.__class__}.")
        return self._transforms


class VirtualFundamentalClassifier(BasicPitchClassifier):
    """ Classifies an event to its corresponding virtual fundamental in range [0, 11].
            Corpus: Uses EventParameter VirtualFundamental in range [128, 139]
            Influence: Responds to keyword "pitch" in range [128, 139]. """

    @staticmethod
    def _label_from_corpus_event(event: CorpusEvent, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(event.get_feature(VirtualFundamental)).value() % 12
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _label_from_feature(pitch: AbstractFeature, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(pitch).value() % 12
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _trait_from_corpus_event(event: CorpusEvent) -> int:
        return event.get_feature(VirtualFundamental).value() % 12

    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :raises TransformError if transform_handler doesn't contain any applicable transforms """
        self._transforms = transform_handler.get_by_feature(AbstractIntegerPitch)
        if not self._transforms:
            raise TransformError(f"No applicable transform exists in classifier {self.__class__}.")
        return self._transforms


class BassNoteClassifier(BasicPitchClassifier):
    @staticmethod
    def _label_from_corpus_event(event: CorpusEvent, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(event.get_feature(BassNote)).value()
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _label_from_feature(pitch: AbstractFeature, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(pitch).value()
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _trait_from_corpus_event(event: CorpusEvent) -> int:
        return event.get_feature(BassNote).value()

    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :raises TransformError if transform_handler doesn't contain any applicable transforms """
        self._transforms = transform_handler.get_by_feature(AbstractIntegerPitch)
        if not self._transforms:
            raise TransformError(f"No applicable transform exists in classifier {self.__class__}.")
        return self._transforms


class BassNoteMod12Classifier(BasicPitchClassifier):
    @staticmethod
    def _label_from_corpus_event(event: CorpusEvent, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(event.get_feature(BassNote)).value() % 12
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _label_from_feature(pitch: AbstractFeature, transform: AbstractTransform) -> IntLabel:
        inverse_transformed_label: int = transform.inverse(pitch).value() % 12
        return IntLabel(inverse_transformed_label)

    @staticmethod
    def _trait_from_corpus_event(event: CorpusEvent) -> int:
        return event.get_feature(BassNote).value() % 12

    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :raises TransformError if transform_handler doesn't contain any applicable transforms """
        self._transforms = transform_handler.get_by_feature(AbstractIntegerPitch)
        if not self._transforms:
            raise TransformError(f"No applicable transform exists in classifier {self.__class__}.")
        return self._transforms



from typing import Any, Dict

from somax.features.feature import CorpusFeature


class MaxVelocity(CorpusFeature):
    def __init__(self, value: float):
        super().__init__(value=value)

    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        for event in corpus.events:
            event.set_feature(cls(value=max([note.velocity for note in event.notes]) / 128))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', _fg_spectrogram: Spectrogram, _bg_spectrogram: Spectrogram,
    #             _fg_chromagram: Chromagram, _bg_chromagram: Chromagram, **kwargs):
    #     return cls(value=max([note.velocity for note in event.notes]) / 128)

    @classmethod
    def decode(cls, trait_dict: Dict[str, Any]) -> 'CorpusFeature':
        return cls(value=trait_dict["velocity"])

    def encode(self) -> Dict[str, Any]:
        return {"velocity": self._value}

    def value(self) -> Any:
        return self._value


class VerticalDensity(CorpusFeature):
    def __init__(self, value: int):
        super().__init__(value=value)

    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        for event in corpus.events:
            event.set_feature(cls(value=len(event.notes)))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', _fg_spectrogram: Spectrogram, _bg_spectrogram: Spectrogram,
    #             _fg_chromagram: Chromagram, _bg_chromagram: Chromagram, **kwargs):
    #     return cls(value=len([event.notes]))

    @classmethod
    def decode(cls, trait_dict: Dict[str, Any]) -> 'CorpusFeature':
        return cls(value=trait_dict["density"])

    def encode(self) -> Dict[str, Any]:
        return {"density": self._value}

    def value(self) -> Any:
        return self._value



import asyncio
import ipaddress
import logging
from abc import ABC, abstractmethod
from logging import Logger
from typing import Optional

from maxosc.caller import Caller


# TODO: This should really be part of MaxOsc package, not Somax.
from maxosc.maxformatter import MaxFormatter
from pythonosc.dispatcher import Dispatcher
from pythonosc.osc_server import AsyncIOOSCUDPServer

from somax import settings
from somax.runtime.target import Target, SimpleOscTarget, SendProtocol


class AsyncioOscObject(Caller, ABC):
    IP_LOCALHOST = "127.0.0.1"
    DEFAULT_CALLBACK_INTERVAL = 0.001

    def __init__(self, recv_port: int, send_port: int, ip: str, address: str, debug: bool = settings.DEBUG):
        super().__init__(parse_parenthesis_as_list=False, discard_duplicate_args=False)
        self.logger = logging.getLogger(__name__)
        self.recv_port: int = recv_port
        self.send_port: int = send_port
        self.ip: str = ip
        self.address: str = address
        self.debug: bool = debug
        self.target: Target = SimpleOscTarget(address, send_port, ip)
        self.server: Optional[AsyncIOOSCUDPServer] = None

    @abstractmethod
    async def _main_loop(self):
        pass

    async def _run(self):
        """ raises: OSError is server already is in use """
        osc_dispatcher: Dispatcher = Dispatcher()
        osc_dispatcher.map(self.address, self.__process_osc)
        osc_dispatcher.set_default_handler(self.__unmatched_osc)
        self.server: AsyncIOOSCUDPServer = AsyncIOOSCUDPServer((self.ip, self.recv_port),
                                                               osc_dispatcher, asyncio.get_event_loop())
        transport, protocol = await self.server.create_serve_endpoint()
        self.target.send(SendProtocol.AGENT_INSTANTIATED, Target.WRAPPED_BANG)
        await self._main_loop()
        transport.close()

    def __process_osc(self, _address, *args):
        args_str: str = MaxFormatter.format_as_string(*args)
        try:
            self.call(args_str)
        except Exception as e:
            self.logger.error(e)
            self.logger.debug(repr(e))
            if self.debug:
                raise

    def __unmatched_osc(self, address: str, *_args, **_kwargs) -> None:
        self.logger.info(f"The address '{address}' does not exist.")




    def parse_ip(self, ip: str, logger: Optional[Logger] = None) -> str:
        try:
            ipaddress.ip_address(ip)
            return ip
        except ValueError as e:
            err = f"{str(e)}. Setting ip to {AsyncioOscObject.IP_LOCALHOST}."
            logger.error(err)
            return AsyncioOscObject.IP_LOCALHOST

    def parse_osc_address(self, string: str) -> str:
        if not string.startswith("/"):
            self.logger.debug(f"OSC address must begin with '/'. Setting OSC address to '/{string}'.")
            return f"/{string}"
        return string



import logging
import multiprocessing
import os
from typing import Tuple, List, Optional, Dict, Any

import pandas as pd

from somax.corpus_builder.chroma_filter import AbstractFilter
from somax.corpus_builder.chromagram import Chromagram
from somax.corpus_builder.matrix_keys import MatrixKeys as Keys
from somax.corpus_builder.note_matrix import NoteMatrix
from somax.corpus_builder.spectrogram import Spectrogram
from somax.features.feature import CorpusFeature
from somax.runtime.corpus import Corpus, ContentType
from somax.runtime.corpus_event import Note, CorpusEvent
from somax.runtime.osc_log_forwarder import OscLogForwarder
from somax.runtime.target import SimpleOscTarget


# TODO: Simple prototype to test the idea of multithreaded corpusbuilding
class ThreadedCorpusBuilder(multiprocessing.Process):
    def __init__(self, filepath: str, corpus_name: Optional[str] = None,
                 foreground_channels: Tuple[int] = tuple(range(1, 17)),
                 background_channels: Tuple[int] = tuple(range(1, 17)),
                 spectrogram_filter: AbstractFilter = AbstractFilter.parse(AbstractFilter.DEFAULT),
                 output_folder: Optional[str] = None, overwrite: bool = False,
                 osc_address: Optional[str] = None, ip: Optional[str] = None,
                 send_port: Optional[int] = None, **kwargs):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.filepath = filepath
        self.corpus_name = corpus_name
        self.foreground_channels = foreground_channels
        self.background_channels = background_channels
        self.spectrogram_filter = spectrogram_filter
        self.output_folder = output_folder
        self.overwrite = overwrite
        self.osc_address = osc_address
        self.ip = ip
        self.send_port = send_port
        self.kwargs = kwargs

    def run(self):
        if not (self.osc_address is None or self.ip is None or self.send_port is None):
            self.logger.addHandler(OscLogForwarder(SimpleOscTarget(address=self.osc_address,
                                                                   port=self.send_port, ip=self.ip)))
            print(self.logger)
            self.logger.setLevel(logging.INFO)

        try:
            corpus: Corpus = CorpusBuilder().build(filepath=self.filepath, corpus_name=self.corpus_name,
                                                   spectrogram_filter=self.spectrogram_filter, **self.kwargs)
            self.logger.debug(f"[build_corpus]: Successfully built '{corpus.name}' from file '{self.filepath}'.")
        except ValueError as e:  # TODO: Missing all exceptions from CorpusBuilder.build()
            self.logger.error(f"{str(e)} No Corpus was built.")
            return

        if self.output_folder is not None:
            self.logger.info(f"[build_corpus]: Exporting corpus '{corpus.name}' to path '{self.output_folder}'...")
            try:
                output_filepath: str = corpus.export(self.output_folder, overwrite=self.overwrite)
                self.logger.info(f"Corpus was successfully written to file '{output_filepath}'.")
            except (IOError, AttributeError, KeyError) as e:
                self.logger.error(f"{str(e)} Export of corpus failed.")


class CorpusBuilder:
    MIDI_FILE_EXTENSIONS = [".mid", ".midi"]
    AUDIO_FILE_EXTENSIONS = []
    CORPUS_FILE_EXTENSIONS = [".json"]

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def build(self, filepath: str, corpus_name: Optional[str] = None,
              foreground_channels: Tuple[int] = tuple(range(1, 17)),
              background_channels: Tuple[int] = tuple(range(1, 17)),
              spectrogram_filter: AbstractFilter = AbstractFilter.parse(AbstractFilter.DEFAULT),
              **kwargs) -> Corpus:
        """ :raises TODO!!! """
        # TODO: Handle folders
        if os.path.isdir(filepath):
            raise NotImplementedError("Building corpora from folders is not supported yet.")
        else:
            filename, extension = os.path.splitext(filepath.split("/")[-1])
            name = corpus_name if corpus_name is not None else filename
            if extension in CorpusBuilder.MIDI_FILE_EXTENSIONS:
                corpus: Corpus = self._build_midi(filepath, name, foreground_channels, background_channels,
                                                  spectrogram_filter, **kwargs)
            elif extension in CorpusBuilder.AUDIO_FILE_EXTENSIONS:
                corpus: Corpus = self._build_audio(filepath, name, foreground_channels, background_channels,
                                                   spectrogram_filter, **kwargs)
            else:
                raise IOError("Invalid file format. Valid extensions are {}.".format(
                    "','".join(self.MIDI_FILE_EXTENSIONS + self.AUDIO_FILE_EXTENSIONS)))
        return corpus

    def _build_midi(self, filepath: str, name: str, foreground_channels: Tuple[int],
                    background_channels: Tuple[int], spectrogram_filter: AbstractFilter, **kwargs) -> Corpus:
        # TODO: Option to plot note matrix, spectrograms, chromagrams and slices along the way!
        self.logger.debug(f"Building midi corpus {name}")
        note_matrix: NoteMatrix = NoteMatrix.from_midi_file(filepath)
        self.logger.debug(f"Note matrix {note_matrix} constructed.")
        fg_matrix: NoteMatrix = note_matrix.split_by_channel(foreground_channels)
        bg_matrix: NoteMatrix = note_matrix.split_by_channel(background_channels)
        fg_spectrogram: Spectrogram = Spectrogram.from_midi(fg_matrix, spectrogram_filter, **kwargs)
        bg_spectrogram: Spectrogram = Spectrogram.from_midi(bg_matrix, spectrogram_filter, **kwargs)
        self.logger.debug(f"Spectrograms {fg_spectrogram} and {bg_spectrogram} constructed.")
        fg_chromagram: Chromagram = Chromagram.from_midi(fg_spectrogram, **kwargs)
        bg_chromagram: Chromagram = Chromagram.from_midi(bg_spectrogram, **kwargs)
        self.logger.debug(f"Chromagrams {fg_chromagram} and {bg_chromagram} constructed.")

        build_parameters: Dict[str, Any] = {"name": name,
                                            "foreground_channels": foreground_channels,
                                            "background_channels": background_channels,
                                            "spectrogram_filter": spectrogram_filter.build_parameters,
                                            "spectrogram_parameters": fg_spectrogram.build_parameters,
                                            "chromagram_parameters": fg_chromagram.build_parameters
                                            }

        corpus: Corpus = self.slice_midi(note_matrix, name, build_parameters=build_parameters, **kwargs)
        corpus.fg_spectrogram = fg_spectrogram
        corpus.bg_spectrogram = bg_spectrogram
        corpus.fg_chromagram = fg_chromagram
        corpus.bg_chromagram = bg_chromagram
        self.logger.debug(f"Corpus {corpus} initialized from slicing note matrix.")

        for _, feature in CorpusFeature.all_corpus_features():
            corpus = feature.analyze(corpus, **kwargs)

        self.logger.debug(f"Analysis of Corpus {corpus} was completed.")

        return corpus

    def _build_audio(self, filepath: str, name: str, *args, **kwargs) -> Corpus:
        pass

    # def build_audio(self):
    #     build_parameters = {...}   # stored in corpus to know how corpus was constructed
    #     audio_data: np.ndarray = AudioParser.read_audio()   # Handle multichannel files too! + meaningful visualization
    #     audio_stft_fg, bg: np.ndarray = Stft.audio_stft()
    #     audio_chromagram_fg, bg: np.ndarray = Chromagram.audio_chromagram()
    #     corpus: Corpus = CorpusBuilder.slice_audio(audio_data, params)
    #     note_matrix: NoteMatrix = PolyF0.estimate(audio_data)
    #     corpus.populate(note_matrix)
    #     for eventparam in eventparams:
    #         eventparam.classify(corpus, audio_stft_fg, bg, audio_chromagram_fg, bg)
    #     return corpus

    def slice_midi(self, note_matrix: NoteMatrix, name: str, build_parameters: Dict[str, Any],
                   slice_tolerance_ms: float = 30.0, **_kwargs) -> Corpus:
        index: int = 0
        events: List[CorpusEvent] = [CorpusEvent.incomplete(index, note_matrix.notes.iloc[0])]
        # TODO: [OPTIMIZATION]Using iterrows will be very slow for large matrices. Subject to optimization
        for i, note in note_matrix.notes.iloc[1:].iterrows():  # type: pd.Series
            # Finalize previous and create a new CorpusEvent
            if note[Keys.ABS_ONSET] > events[-1].absolute_onset + slice_tolerance_ms:
                events[-1].set_duration(end=note[Keys.REL_ONSET], absolute_end=note[Keys.ABS_ONSET])
                index += 1
                corpus_event: CorpusEvent = CorpusEvent.incomplete(index, note)
                corpus_event.append_notes(events[-1].held_from(), events[-1].onset, events[-1].absolute_onset)
                events.append(corpus_event)
            # Append to previous Corpus Event
            else:
                events[-1].append_raw(note)

        # Finalize last slice
        final_note: Note = max(events[-1].notes, key=lambda n: n.onset + n.duration)
        note_end: float = events[-1].onset + max(0.0, final_note.duration + final_note.onset)
        note_absolute_end: float = events[-1].absolute_onset \
                                   + max(0.0, final_note.absolute_duration + final_note.absolute_onset)
        events[-1].set_duration(end=note_end, absolute_end=note_absolute_end)

        build_parameters["slice_tolerance_ms"] = slice_tolerance_ms

        return Corpus(events, name, ContentType.MIDI, build_parameters=build_parameters)



from abc import ABC
from enum import Enum
from typing import Any

from somax.features.feature import RuntimeFeature
from somax.runtime.corpus_event import CorpusEvent


class InfluenceType(Enum):
    PITCH = "pitch"
    CHROMA = "chroma"

    @classmethod
    def from_string(cls, influence_type: str):
        """ :raises ValueError """
        return InfluenceType(influence_type.lower())


class AbstractInfluence(ABC):
    pass


class CorpusInfluence(AbstractInfluence):
    def __init__(self, corpus_event: CorpusEvent):
        self.corpus_event: CorpusEvent = corpus_event


class FeatureInfluence(AbstractInfluence):
    def __init__(self, feature: RuntimeFeature):
        self.feature = feature

    @classmethod
    def from_keyword(cls, keyword: str, value: Any):
        """ :raises ValueError if a feature matching the keyword doesn't exist """
        return cls(RuntimeFeature.from_string(keyword, value))

# TODO: REMOVE
# class KeywordInfluence(AbstractInfluence):
#     def __init__(self, keyword: InfluenceType, influence_data: Any):
#         self.keyword: InfluenceType = keyword
#         self.influence_data: Any = influence_data



from typing import Dict, Any

import numpy as np

from somax.settings import IMPORT_MATPLOTLIB

if IMPORT_MATPLOTLIB:
    pass

from .spectrogram import Spectrogram


# TODO: Implement visualization

class Chromagram:
    def __init__(self, chromagram: np.ndarray, duration_ms: float):
        self.chromagram: np.ndarray = chromagram
        self.duration_ms: float = duration_ms
        self.time_factor: float = 1 / duration_ms * chromagram.shape[1]

    @classmethod
    def from_midi(cls, spectrogram: Spectrogram, **_kwargs):
        raw_spectrogram: np.ndarray = spectrogram.spectrogram
        chromagram: np.ndarray = np.zeros((12, raw_spectrogram.shape[1]))
        for i in range(128):
            pitch_class = i % 12
            chromagram[pitch_class, :] += raw_spectrogram[i, :]
        return cls(chromagram, spectrogram.duration_ms)

    def at(self, onset_ms: float) -> np.ndarray:
        return self.chromagram[:, int(np.floor(onset_ms * self.time_factor))]

    # TODO: Removed for PyInstaller to exclude matplotlib
    # def plot(self, ax: Optional[Axes] = None):
    #     if not ax:
    #         fig = plt.figure()
    #         ax: Axes = fig.add_subplot(1, 1, 1)
    #
    #     ax.imshow(self.chromagram, aspect='auto', origin='lower', norm=colors.LogNorm(vmin=0.0001, vmax=20),
    #               extent=[0, self.duration_ms, 0, 128])
    #     ax.set_ylabel("Note Number")
    #     ax.set_xlabel("Time [ms]")
    #     # plt.colorbar()

    @property
    def build_parameters(self) -> Dict[str, Any]:
        return {}



from abc import ABC
from importlib import resources
from typing import List, Tuple, Optional

import numpy as np
from sklearn.mixture import GaussianMixture

from somax.classification import tables
from somax.classification.classifier import AbstractClassifier
from somax.features import BackgroundChroma
from somax.features.feature import RuntimeFeature, CorpusFeature
from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import InvalidLabelInput, TransformError, InvalidCorpus
from somax.runtime.influence import AbstractInfluence, CorpusInfluence, FeatureInfluence
from somax.runtime.label import AbstractLabel, IntLabel
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform


# TODO: Normalization. Normalization has been removed for now, but should be thoroughly tested.

# TODO: This class needs a lot of optimization. A larger corpus will mean higher computation time to a
#  point where classification of an influence might not be real-time anymore.


class ChromaClassifier(AbstractClassifier, ABC):

    def update_transforms(self, transform_handler: TransformHandler) -> List[AbstractTransform]:
        """ :raises TransformError if transform_handler doesn't contain any applicable transforms """
        self._transforms = transform_handler.get_by_feature(BackgroundChroma)
        if not self._transforms:
            raise TransformError(f"No applicable transform exists in classifier {self.__class__}.")
        return self._transforms


class SomChromaClassifier(ChromaClassifier):
    """ Classifies an event according to its chroma based on a pre-computed self-organizing map.
                Corpus:    Uses foreground chroma (12 non-normalized floats) of EventParameter OnsetChroma
                Influence: Responds to keyword "chroma" followed by 12 non-normalized floats. """

    SOM_DATA_FILE = 'misc_hsom'  # Note: vectors in file are not normalized
    SOM_CLASS_FILE = 'misc_hsom_c'

    USE_MULTIPROCESSING = True

    def __init__(self):
        super().__init__()
        with resources.path(tables, self.SOM_DATA_FILE) as path:
            self._som_data = np.loadtxt(path.absolute(), dtype=np.float32, delimiter=",")  # Shape: (N, 12)
        with resources.path(tables, self.SOM_CLASS_FILE) as path:
            self._som_classes = np.loadtxt(path.absolute(), dtype=int, delimiter=",")  # Shape: (N,)

    def cluster(self, corpus: Corpus, **kwargs) -> None:
        # No clustering required for class
        pass

    def classify_corpus(self, corpus: Corpus) -> List[IntLabel]:
        if self.USE_MULTIPROCESSING:
            import multiprocessing
            with multiprocessing.Pool(processes=4) as pool:
                labels: List[IntLabel] = pool.map(self._multiproc_compute_label, corpus.events)
        else:
            labels: List[IntLabel] = []
            for event in corpus.events:  # type: CorpusEvent
                labels.append(self._label_from_chroma(event.get_feature(BackgroundChroma).value()))
        return labels

    def _multiproc_compute_label(self, e: CorpusEvent):
        return self._label_from_chroma(e.get_feature(BackgroundChroma).value())

    def classify_influence(self, influence: AbstractInfluence) -> List[Tuple[AbstractLabel, AbstractTransform]]:
        """ :raises TransformError if no transforms exist """
        if not self._transforms:  # transforms is empty
            raise TransformError(f"No transforms exist in classifier {self}")
        if isinstance(influence, FeatureInfluence) and isinstance(influence.feature, BackgroundChroma):
            chroma: RuntimeFeature = influence.feature
            return [((self._label_from_chroma(t.inverse(chroma).value())), t)
                    for t in self._transforms]
        elif isinstance(influence, CorpusInfluence):
            chroma: CorpusFeature = influence.corpus_event.get_feature(BackgroundChroma)
            return [(self._label_from_chroma(t.inverse(chroma).value()), t) for t in self._transforms]
        else:
            raise InvalidLabelInput(f"Influence {influence} could not be classified by {self}.")

    def _label_from_chroma(self, chroma: np.ndarray) -> IntLabel:
        # max_val: float = np.max(chroma)
        # if max_val > 0:
        #     chroma /= max_val
        rms: np.ndarray = np.sqrt(np.sum(np.power(chroma - self._som_data, 2), axis=1))
        return IntLabel(self._som_classes[np.argmin(rms)])

    def clear(self) -> None:
        pass  # SomChromaClassifier is stateless


class GmmClassifier(ChromaClassifier, ABC):
    USE_MULTIPROCESSING = True

    def __init__(self, num_components: int = 100, max_iter: int = 500):
        super().__init__()
        self.num_components: int = num_components
        self.max_iter: int = max_iter
        self.gmm: Optional[GaussianMixture] = None

    def classify_corpus(self, corpus: Corpus) -> List[AbstractLabel]:
        if self.USE_MULTIPROCESSING:
            import multiprocessing
            with multiprocessing.Pool(processes=4) as pool:
                labels: List[IntLabel] = pool.map(self._multiproc_compute_label, corpus.events)
        else:
            labels: List[IntLabel] = []
            for event in corpus.events:  # type: CorpusEvent
                chroma: np.ndarray = event.get_feature(BackgroundChroma).value().reshape(1, -1)
                # max_val: float = np.max(chroma)
                # if max_val > 0:
                #     chroma /= max_val
                labels.append(IntLabel(int(self.gmm.predict(chroma))))
        return labels

    def _multiproc_compute_label(self, e: CorpusEvent) -> IntLabel:
        return IntLabel(int(self.gmm.predict(e.get_feature(BackgroundChroma).value().reshape(1, -1))))

    def classify_influence(self, influence: AbstractInfluence) -> List[Tuple[AbstractLabel, AbstractTransform]]:
        """ :raises TransformError if no transforms exist """
        if not self._transforms:
            raise TransformError(f"No Transforms exist in classifier {self}")
        if isinstance(influence, FeatureInfluence) and isinstance(influence.feature, BackgroundChroma):
            chroma: RuntimeFeature = influence.feature
        elif isinstance(influence, CorpusInfluence):
            chroma: CorpusFeature = influence.corpus_event.get_feature(BackgroundChroma)
        else:
            raise InvalidLabelInput(f"Influence {influence} could not be classified by {self}.")

        # max_val: float = np.max(chroma)
        # if max_val > 0:
        #     chroma /= max_val
        return [(IntLabel(int(self.gmm.predict(t.inverse(chroma).value().reshape(1, -1)))), t)
                for t in self._transforms]

    def clear(self) -> None:
        pass  # GmmClassifier is stateless


class AbsoluteGmmClassifier(GmmClassifier):
    GMM_DATA_FILE = 'misc_hsom'

    def __init__(self, num_components: int = 100, max_iter: int = 500):
        super().__init__(num_components, max_iter)
        with resources.path(tables, self.GMM_DATA_FILE) as p:
            self._gmm_data: np.ndarray = np.loadtxt(p.absolute(), dtype=np.float32, delimiter=",")  # Shape: (N, 12)

    def cluster(self, _corpus: Corpus) -> None:
        self.gmm = GaussianMixture(n_components=self.num_components, max_iter=self.max_iter).fit(self._gmm_data)


class RelativeGmmClassifier(GmmClassifier):

    def cluster(self, corpus: Corpus) -> None:
        """ :raises InvalidCorpus if number of events in corpus is lower than `self.num_components`."""
        chromas: List[np.ndarray] = [event.get_feature(BackgroundChroma).value() for event in corpus.events]
        gmm_data: np.ndarray = np.row_stack(chromas)
        max_per_col: np.ndarray = np.max(chromas, axis=1)
        max_per_col[max_per_col == 0] = 1  # don't normalize empty vectors - avoid div0 error
        chromas /= max_per_col[:, np.newaxis]
        try:
            self.gmm = GaussianMixture(n_components=self.num_components, max_iter=self.max_iter).fit(gmm_data)
        except ValueError as e:
            if self.num_components > corpus.length():
                raise InvalidCorpus(f"{self.__class__.__name__} could not classify corpus '{str(corpus)}' since corpus "
                                    f"length ({corpus.length()}) is lower than number of requested clusters "
                                    f"({self.num_components}). "
                                    f"Reduce the number of clusters or select another classifier")
            else:
                raise InvalidCorpus(f"Unknown error encountered in {self.__class__.__name__}. Error: {repr(e)}.")




import logging
from typing import Dict, Optional, Tuple, Type, List

import numpy as np

from somax.runtime.atom import Atom
from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import DuplicateKeyError
from somax.runtime.exceptions import InvalidCorpus, InvalidLabelInput
from somax.runtime.improvisation_memory import ImprovisationMemory
from somax.runtime.influence import AbstractInfluence
from somax.runtime.merge_actions import AbstractMergeAction
from somax.runtime.peak_selector import AbstractPeakSelector
from somax.runtime.peaks import Peaks
from somax.runtime.scale_actions import AbstractScaleAction
from somax.runtime.streamview import Streamview
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform, NoTransform
from somax.scheduler.scheduled_object import ScheduledMidiAgent, TriggerMode


class Player(Streamview, ScheduledMidiAgent):

    def __init__(self, name: str,
                 trigger_mode: TriggerMode = TriggerMode.default(),
                 peak_selector: AbstractPeakSelector = AbstractPeakSelector.default(),
                 merge_action: AbstractMergeAction = AbstractMergeAction.default(),
                 corpus: Optional[Corpus] = None,
                 scale_actions: List[AbstractScaleAction] = AbstractScaleAction.default_set(),
                 **kwargs):
        self._transform_handler: TransformHandler = TransformHandler()
        super().__init__(name, transform_handler=self._transform_handler, corpus=corpus, merge_action=merge_action,
                         trigger_mode=trigger_mode, **kwargs)
        self.logger = logging.getLogger(__name__)
        self.peak_selector: AbstractPeakSelector = peak_selector
        self.corpus: Optional[Corpus] = corpus
        self.scale_actions: Dict[Type[AbstractScaleAction], AbstractScaleAction] = {}

        for scale_action in scale_actions:
            self.add_scale_action(scale_action)

        self.improvisation_memory: ImprovisationMemory = ImprovisationMemory()
        self.previous_peaks: Peaks = Peaks.create_empty()
        self._transform_handler: TransformHandler = TransformHandler()

        self._force_jump_index: Optional[int] = None

        self._parse_parameters()

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name}, ...)"

    ######################################################
    # MAIN RUNTIME FUNCTIONS
    ######################################################

    def new_event(self, scheduler_time: float, tempo: float) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        self.logger.debug(f"[new_event] Player '{self.name}' attempting to create a new event "
                          f"at scheduler time '{scheduler_time}'.")
        if not bool(self.enabled.value):
            return None
        if not self.corpus:
            raise InvalidCorpus(f"No Corpus has been loaded in player '{self.name}'.")

        if self._force_jump_index is not None:
            self.clear()
            event_and_transform: Optional[Tuple[CorpusEvent, AbstractTransform]]
            event_and_transform = self._force_jump()
        else:
            self._update_peaks_on_new_event(scheduler_time)
            peaks: Peaks = self._merged_peaks(scheduler_time, self.improvisation_memory, self.corpus)
            peaks = self._scale_peaks(peaks, scheduler_time, self.improvisation_memory, self.corpus)

            event_and_transform: Optional[Tuple[CorpusEvent, AbstractTransform]]
            event_and_transform = self.peak_selector.decide(peaks, self.improvisation_memory,
                                                            self.corpus, self._transform_handler)
            self.previous_peaks = peaks
        if event_and_transform is None:
            self._feedback(None, scheduler_time, NoTransform())
            return None
        event, transform = event_and_transform
        event = transform.apply(event)  # returns deepcopy of transformed event

        self.improvisation_memory.append(event, scheduler_time, transform, tempo,
                                         artificially_sustained=self.hold_notes_artificially,
                                         simultaneous_onsets=self.simultaneous_onsets)
        self._feedback(event, scheduler_time, transform)
        return event, transform

    def influence(self, path: List[str], influence: AbstractInfluence, time: float, **kwargs) -> Dict[Atom, int]:
        """ Raises: InvalidLabelInput (if influencing a specific path without matching label), KeyError
            Return values are only for gathering statistics (Evaluator, etc.) and not used in runtime."""
        if not bool(self.enabled.value):
            return {}
        if self.corpus is None:
            raise InvalidCorpus(f"No Corpus has been loaded in player '{self.name}'.")
        num_generated_peaks: Dict[Atom, int] = {}
        if not path:
            for atom in self._direct_influenced_atoms():
                try:
                    num_peaks: int = atom.influence(influence, time, **kwargs)
                    num_generated_peaks[atom] = num_peaks
                except InvalidLabelInput:
                    # Ignore atom if label doesn't match
                    continue
        else:
            atom: Atom = self._get_atom(path)
            num_peaks: int = atom.influence(influence, time, **kwargs)
            num_generated_peaks[atom] = num_peaks
        return num_generated_peaks

    ######################################################
    # MODIFY STATE
    ######################################################

    def _feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                  applied_transform: AbstractTransform) -> None:
        self.peak_selector.feedback(feedback_event, time, applied_transform)
        for scale_action in self.scale_actions.values():
            scale_action.feedback(feedback_event, time, applied_transform)
        self.feedback(feedback_event, time, applied_transform)

    def clear(self):
        """ Reset runtime state without modifying any parameters or settings """
        self.previous_peaks = Peaks.create_empty()
        for scale_action in self.scale_actions.values():
            scale_action.clear()
        Streamview.clear(self)
        self._transform_handler.clear()

    def clear_memory(self):
        self.improvisation_memory = ImprovisationMemory()

    def force_jump(self, index: int):
        self._force_jump_index = index

    def read_corpus(self, corpus: Corpus) -> None:
        self._update_transforms()
        self.corpus = corpus
        Streamview.read_corpus(self, corpus)

    def set_peak_selector(self, peak_selector: AbstractPeakSelector) -> None:
        self.peak_selector = peak_selector
        self._parse_parameters()

    def add_scale_action(self, scale_action: AbstractScaleAction, override: bool = False):
        """ Raises: DuplicateKeyError """
        if type(scale_action) in self.scale_actions and not override:
            raise DuplicateKeyError(f"A Scale Action of type '{type(scale_action).__name__}' already exists."
                                    f"To override: use 'override=True'.")
        else:
            scale_action.update_transforms(self._transform_handler)
            self.scale_actions[type(scale_action)] = scale_action

            self._parse_parameters()

    def remove_scale_action(self, scale_action_type: Type[AbstractScaleAction]):
        """ Raises: KeyError """
        del self.scale_actions[scale_action_type]
        del self.parameter_dict[scale_action_type.__name__]

    def add_transform(self, transform: AbstractTransform):
        """ :raises TransformError if a transform of the same instance with the same parameters already exists """
        self._transform_handler.add(transform)
        self._update_transforms()

    def remove_transform(self, transform: AbstractTransform):
        """ :raises IndexError if key doesn't exist
                    TransformError if attempting to delete last transform
        """
        self._transform_handler.remove(transform)
        self._update_transforms()

    def export_runtime_corpus(self, name: str, **kwargs) -> Corpus:
        """ raises: InvalidCorpus if there's no data to export"""
        return self.improvisation_memory.export(name, self.corpus, **kwargs)

    def get_peaks(self) -> Tuple[Dict[str, int], int]:
        peaks_count: Dict[str, int] = {self.name: self.previous_peaks.size()}
        for atom in self.all_atoms():
            peaks: Peaks = atom.get_peaks()
            peaks_count[atom.name] = peaks.size()
        return peaks_count, self.improvisation_memory.length()

    ######################################################
    # PRIVATE
    ######################################################

    def _force_jump(self) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        self.clear()
        index: Optional[int] = self._force_jump_index
        self._force_jump_index = None
        try:
            event: CorpusEvent = self.corpus.events[index]
            return event, NoTransform()
        except (IndexError, TypeError, AttributeError) as e:
            self.logger.debug(f"[_force_jump]: Force jump cancelled due to error: {repr(e)}")
            print(f"[_force_jump]: Force jump cancelled due to error: {repr(e)}")
            return None

    def _scale_peaks(self, peaks: Peaks, scheduler_time: float, influence_history: ImprovisationMemory,
                     corpus: Corpus, **kwargs):
        if peaks.is_empty():
            return peaks
        corresponding_events: List[CorpusEvent] = corpus.events_around(peaks.times)
        corresponding_transforms: List[AbstractTransform] = [self._transform_handler.get_transform(t)
                                                             for t in np.unique(peaks.transform_ids)]
        for scale_action in self.scale_actions.values():
            if scale_action.is_enabled():
                peaks = scale_action.scale(peaks, scheduler_time, corresponding_events, corresponding_transforms,
                                           influence_history, corpus, **kwargs)
        return peaks

    def _update_transforms(self):
        for scale_action in self.scale_actions.values():
            scale_action.update_transforms(self._transform_handler)
        super().update_transforms()



import copy
from typing import Tuple, List

from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.exceptions import InvalidCorpus
from somax.runtime.memory_state import MemoryState
from somax.runtime.transforms import AbstractTransform


class ImprovisationMemory:

    def __init__(self):
        self._history: List[Tuple[CorpusEvent, MemoryState]] = []

    def append(self, event: CorpusEvent, trigger_time: float, transforms: AbstractTransform, tempo: float,
               artificially_sustained: bool, simultaneous_onsets: bool) -> None:
        """ Note: Transform has already been applied to the event, which is a deepcopy of the original event
                  (see `Player.new_event()` for more info)"""
        self._history.append((event, MemoryState(trigger_time, transforms, tempo,
                                                 artificially_sustained, simultaneous_onsets)))

    def get(self, index: int) -> Tuple[CorpusEvent, float, AbstractTransform]:
        # TODO: Handle with memory state class
        """ raises: IndexError if index is out of range """
        v: Tuple[CorpusEvent, MemoryState] = self._history[index]
        return v[0], v[1].trigger_time, v[1].applied_transform

    def get_latest(self) -> Tuple[CorpusEvent, float, AbstractTransform]:
        """ raises: IndexError if history is empty """
        # TODO: Handle with memory state class
        v: Tuple[CorpusEvent, MemoryState] = self._history[len(self._history) - 1]
        return v[0], v[1].trigger_time, v[1].applied_transform

    def length(self) -> int:
        return len(self._history)

    def export(self, name: str, source_corpus: Corpus, use_original_tempo: bool = False) -> Corpus:
        """ raises: InvalidCorpus if there is no data to export
            TODO: `use_original_corpus` is a temporary workaround to handle the tempo offset described
                  in https://trello.com/c/vKfkisIV. Remove this once a proper solution is in place.
        """
        if len(self._history) == 0:
            raise InvalidCorpus("The recorded history is empty")
        elapsed_abs_time: float = 0.0
        events: List[CorpusEvent] = []
        for event, memory_state in copy.deepcopy(self._history):  # type: CorpusEvent, MemoryState
            current_onset: float = memory_state.trigger_time
            if use_original_tempo:
                memory_state.tempo = event.tempo
            if len(events) > 0:
                elapsed_time: float = current_onset - events[-1].onset
                events[-1].duration = elapsed_time
                elapsed_abs_time += elapsed_time * 60 / memory_state.tempo
            event.onset = current_onset
            event.absolute_onset = elapsed_abs_time
            event.recorded_memory_state = memory_state
            events.append(event)
        return Corpus(events, name, content_type=source_corpus.content_type,
                      build_parameters={"build_method": "runtime"})

    def get_n_latest(self, n: int) -> List[Tuple[CorpusEvent, float, AbstractTransform]]:
        """ :returns n latest events in reverse order (index 0 is latest event)"""
        if len(self._history) < n:
            return [(v[0], v[1].trigger_time, v[1].applied_transform) for v in self._history]
        else:
            return [(self._history[-i][0], self._history[-i][1].trigger_time, self._history[-i][1].applied_transform)
                    for i in range(1, n + 1)]



class InvalidPath(Exception):
    def __init__(self, message):
        super(InvalidPath, self).__init__(message)


class InvalidJsonFormat(Exception):
    def __init__(self, error):
        super(InvalidJsonFormat, self).__init__(error)


class InvalidLabelInput(Exception):
    def __init__(self, error):
        super(InvalidLabelInput, self).__init__(error)


class InvalidCorpus(Exception):
    def __init__(self, error):
        super(InvalidCorpus, self).__init__(error)


class TransformError(Exception):
    def __init__(self, error):
        super(TransformError, self).__init__(error)


class TransformIdentityError(Exception):
    def __init__(self, error):
        super().__init__(error)


class InvalidConfiguration(Exception):
    def __init__(self, error):
        super(InvalidConfiguration, self).__init__(error)


class DuplicateKeyError(Exception):
    def __init__(self, error):
        super(DuplicateKeyError, self).__init__(error)


class ParameterError(Exception):
    def __init__(self, error):
        super(ParameterError, self).__init__(error)



from typing import List, Optional, Union

import numpy as np


class Peaks:

    def __init__(self, scores: np.ndarray, times: np.ndarray, transform_hashes: np.ndarray):
        """ scores, times and transform_hashes should have the same sizes. TODO: Assert if performance allows"""
        self.scores: np.ndarray = scores  # shape: 1d, dtype: float
        self.times: np.ndarray = times  # shape: 1d, dtype: float
        self.transform_ids: np.ndarray = transform_hashes  # shape: 1d, dtype: int

    def __repr__(self):
        return f"{type(self).__name__}(scores=<{type(self.scores).__name__},shape={self.scores.shape}>," \
               f"times=<{type(self.times).__name__},shape={self.times.shape}>," \
               f"transform_hashes=<{type(self.transform_ids).__name__},shape={self.transform_ids.shape}>)"

    @classmethod
    def create_empty(cls):
        return cls(np.empty(0, dtype=np.float), np.empty(0, dtype=np.float), np.empty(0, dtype=np.int32))

    @classmethod
    def optimized_copy(cls, other: 'Peaks') -> 'Peaks':
        return cls(np.copy(other.scores), other.times, other.transform_ids)

    @classmethod
    def concatenate(cls, peaks: List['Peaks']):
        if not peaks:
            return Peaks.create_empty()
        scores: np.ndarray = np.concatenate([peak.scores for peak in peaks])
        times: np.ndarray = np.concatenate([peak.times for peak in peaks])
        transform_hashes: np.ndarray = np.concatenate([peak.transform_ids for peak in peaks])
        return cls(scores, times, transform_hashes)

    def append(self, scores: List[float], times: List[float], transform_hashes: List[int]):
        self.scores = np.concatenate((self.scores, scores))
        self.times = np.concatenate((self.times, times))
        self.transform_ids = np.concatenate((self.transform_ids, transform_hashes))

    def remove(self, indices: np.ndarray):
        """ mask: 1d boolean array with same size as scores/times/transforms """
        self.scores = np.delete(self.scores, indices)
        self.times = np.delete(self.times, indices)
        self.transform_ids = np.delete(self.transform_ids, indices)

    def scale(self, factor: Union[float, np.ndarray], index_map: Optional[np.ndarray] = None) -> None:
        """ factor:    single value or np.ndarray of same shape as self.scores

            index_map: boolean array corresponding to the size of self.scores.
                       Uniform scaling if no index map is provided"""
        if index_map is None:
            self.scores *= factor
        else:
            self.scores[index_map] *= factor

    def size(self) -> int:
        return self.scores.size

    def reorder(self, indices: np.ndarray):
        self.scores = self.scores[indices]
        self.times = self.times[indices]
        self.transform_ids = self.transform_ids[indices]

    def is_empty(self) -> bool:
        return self.scores.size == 0

    def dump(self) -> (np.ndarray, np.ndarray, np.ndarray):
        return self.scores, self.times, self.transform_ids



import logging
import random
from abc import abstractmethod, ABC
from typing import Tuple, Optional, List

import numpy as np

from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.improvisation_memory import ImprovisationMemory
from somax.runtime.parameter import Parametric, Parameter
from somax.runtime.peaks import Peaks
from somax.runtime.transforms import AbstractTransform, NoTransform
from somax.runtime.transform_handler import TransformHandler
from somax.utils.introspective import StringParsed


class AbstractPeakSelector(Parametric, StringParsed, ABC):
    def __init__(self, **kwargs):
        super(AbstractPeakSelector, self).__init__()
        self.logger = logging.getLogger(__name__)

    def __repr__(self):
        return f"{type(self).__name__}(...)"

    @abstractmethod
    def _decide_default(self, peaks: Peaks, influence_history: ImprovisationMemory,
                        corpus: Corpus, transform_handler: TransformHandler,
                        **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        """ Default action to perform to select output """

    @abstractmethod
    def _decide_fallback(self, peaks: Peaks, influence_history: ImprovisationMemory,
                         corpus: Corpus, transform_handler: TransformHandler,
                         **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        """ Action to perform if no valid peak exists after calling `_decide_default`"""

    @abstractmethod
    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        """ """

    @classmethod
    def default(cls, **kwargs) -> 'AbstractPeakSelector':
        return MaxPeakSelector()

    @classmethod
    def from_string(cls, peak_selector: str, **kwargs) -> 'AbstractPeakSelector':
        return cls._from_string(peak_selector, **kwargs)

    def decide(self, peaks: Peaks, influence_history: ImprovisationMemory, corpus: Corpus,
               transform_handler: TransformHandler, **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        output: Optional[Tuple[CorpusEvent, AbstractTransform]]
        output = self._decide_default(peaks, influence_history, corpus, transform_handler, **kwargs)
        if output is None:
            output = self._decide_fallback(peaks, influence_history, corpus, transform_handler, **kwargs)
        return output


class AbstractFallbackPeakSelector(AbstractPeakSelector, ABC):
    def _decide_fallback(self, peaks: Peaks, influence_history: ImprovisationMemory, corpus: Corpus,
                         transform_handler: TransformHandler,
                         **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        self.logger.debug("[decide] _decide_fallback called.")
        try:
            last_event, _, last_transform = influence_history.get_latest()
            next_state_idx: int = (last_event.state_index + 1) % corpus.length()
            return corpus.event_at(next_state_idx), last_transform
        except IndexError:
            # If history is empty: play the first event in the corpus
            return corpus.event_at(0), NoTransform()


class MaxPeakSelector(AbstractFallbackPeakSelector):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _decide_default(self, peaks: Peaks, influence_history: ImprovisationMemory,
                        corpus: Corpus, transform_handler: TransformHandler,
                        **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        self.logger.debug("[decide] _decide_default called.")
        if peaks.is_empty():
            return None
        max_peak_value: float = np.max(peaks.scores)
        self.logger.debug(f"[decide_default] Max peak value is {max_peak_value}.")
        max_peaks_idx: List[int] = np.argwhere(np.abs(peaks.scores - max_peak_value) < 0.001)
        peak_idx: int = random.choice(max_peaks_idx)
        transform_hash: int = int(peaks.transform_ids[peak_idx])
        return corpus.event_around(peaks.times[peak_idx]), transform_handler.get_transform(transform_hash)

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass


class ThresholdPeakSelector(MaxPeakSelector):
    DEFAULT_THRESHOLD = 0.1

    def __init__(self, threshold: float = DEFAULT_THRESHOLD, **kwargs):
        super().__init__(**kwargs)
        self._threshold: Parameter = Parameter(threshold, 0, None, "float", "TODOOO")

    def _decide_default(self, peaks: Peaks, influence_history: ImprovisationMemory,
                        corpus: Corpus, transform_handler: TransformHandler,
                        **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        if peaks.is_empty():
            return None
        max_peak_value: float = np.max(peaks.scores)
        if max_peak_value < self.threshold:
            return None
        else:
            return super()._decide_default(peaks, influence_history, corpus, transform_handler)

    def _decide_fallback(self, peaks: Peaks, influence_history: ImprovisationMemory,
                         corpus: Corpus, transform_handler: TransformHandler,
                         **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        return None

    def feedback(self, _feedback_event: Optional[CorpusEvent], _time: float,
                 _applied_transform: AbstractTransform) -> None:
        pass

    @property
    def threshold(self):
        return self._threshold.value


class ProbabilisticPeakSelector(AbstractFallbackPeakSelector):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _decide_default(self, peaks: Peaks, influence_history: ImprovisationMemory,
                        corpus: Corpus, transform_handler: TransformHandler,
                        **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        if peaks.is_empty():
            return None
        score_cumsum: np.ndarray = np.cumsum(peaks.scores)
        max_value: float = score_cumsum[-1] - 1e-5  # slight offset to avoid an extremely rare case of a fp error
        output_target_score: float = float(np.random.random(1) * max_value)
        peak_idx: int = np.argwhere(score_cumsum > output_target_score)[0]
        transform_hash: int = int(peaks.transform_ids[peak_idx])
        return corpus.event_around(peaks.times[peak_idx]), transform_handler.get_transform(transform_hash)

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float,
                 applied_transform: AbstractTransform) -> None:
        pass


class ThresholdProbabilisticPeakSelector(ProbabilisticPeakSelector):
    DEFAULT_THRESHOLD = 0.1

    def __init__(self, threshold: float = DEFAULT_THRESHOLD, **kwargs):
        super().__init__(**kwargs)
        self._threshold: Parameter = Parameter(threshold, 0, None, "float", "TODOOO")

    def _decide_default(self, peaks: Peaks, influence_history: ImprovisationMemory, corpus: Corpus,
                        transform_handler: TransformHandler,
                        **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        if peaks.is_empty():
            return None
        max_peak_value: float = np.max(peaks.scores)
        if max_peak_value < self.threshold:
            return None
        else:
            return super()._decide_default(peaks, influence_history, corpus, transform_handler)

    def _decide_fallback(self, peaks: Peaks, influence_history: ImprovisationMemory, corpus: Corpus,
                         transform_handler: TransformHandler,
                         **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        return None

    @property
    def threshold(self):
        return self._threshold.value


class SparsityProbabilisticPeakSelector(MaxPeakSelector):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.output_probability: Parameter = Parameter(1.0, 0.0, 1.0, "float", "Probability to trigger output")

    def _decide_default(self, peaks: Peaks, influence_history: ImprovisationMemory,
                        corpus: Corpus, transform_handler: TransformHandler,
                        **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        if random.random() > self.output_probability.value:
            return super()._decide_default(peaks, influence_history, corpus, transform_handler, **kwargs)
        else:
            return None

    def _decide_fallback(self, peaks: Peaks, influence_history: ImprovisationMemory,
                         corpus: Corpus, transform_handler: TransformHandler,
                         **kwargs) -> Optional[Tuple[CorpusEvent, AbstractTransform]]:
        return None



from typing import Any, Dict, List

import numpy as np

from somax.features.feature import CorpusFeature


class OctaveBands(CorpusFeature):
    NUM_BANDS = 11

    def __init__(self, value: np.ndarray):
        super().__init__(value=value)

    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        for event in corpus.events:
            band_distribution: np.ndarray = np.zeros(OctaveBands.NUM_BANDS, dtype=float)
            octaves: List[int] = [int(note.pitch // 12) for note in event.notes]
            for octave in octaves:
                band_distribution[octave] += 1.0
            max_val: float = np.max(band_distribution)
            if max_val > 0:
                band_distribution /= max_val
            event.set_feature(cls(value=band_distribution))
        return corpus

    @classmethod
    def decode(cls, trait_dict: Dict[str, Any]) -> 'CorpusFeature':
        return cls(value=trait_dict["band"])

    def encode(self) -> Dict[str, Any]:
        return {"band": self._value.tolist()}

    def value(self) -> np.ndarray:
        return self._value



from typing import Dict, List, Type

from somax.features.feature import AbstractFeature
from somax.runtime.exceptions import TransformError
from somax.runtime.transforms import AbstractTransform, RedundantTransform


class TransformHandler:
    def __init__(self):
        default_transform: AbstractTransform = AbstractTransform.default()
        self._transforms: Dict[int, AbstractTransform] = {0: default_transform}
        self._next_id: int = 1

    def add(self, transform: AbstractTransform):
        """ :raises TransformError if a transform of the same instance with the same parameters already exists """
        if transform in self._transforms.values():
            raise TransformError("A transform with the given parameters given already exists")
        self._transforms[self._next_id] = transform
        self._next_id += 1

    def remove(self, transform: AbstractTransform):
        """ :raises IndexError if key doesn't exist
                    TransformError if attempting to delete last transform
        """
        if self.num_transforms() <= 1:
            raise TransformError("Cannot delete the last transform")
        elif transform not in self._transforms.values():
            raise TransformError("A transform with the given parameters does not exist")
        else:
            # self._transforms = {k: v for (k, v) in self._transforms.items() if v != transform}
            for k, t in self._transforms.items():
                if t == transform:
                    self._transforms[k] = RedundantTransform(t)  # flag transform for deletion without removing it

    def num_transforms(self) -> int:
        return len([t for t in self._transforms.values() if not isinstance(t, RedundantTransform)])

    def get_transform(self, transform_id: int) -> AbstractTransform:
        """ :raises IndexError if key doesn't exist """
        return self._transforms[transform_id]

    def get_id(self, transform: AbstractTransform) -> int:
        for k, v in self._transforms.items():
            if v == transform:
                return k

    def get_by_feature(self, feature: Type[AbstractFeature]) -> List[AbstractTransform]:
        transforms: List[AbstractTransform] = []
        for transform in self._transforms.values():
            if not isinstance(transform, RedundantTransform):
                for transform_feature in transform.valid_features():
                    if issubclass(feature, transform_feature):
                        transforms.append(transform)
        return transforms
        # return [t for t in self._transforms.values() if feature in t.valid_features()]

    def clear(self):
        self._transforms = {k: t for (k, t) in self._transforms.items() if not isinstance(t, RedundantTransform)}



from math import log, ceil, floor

def hindemith(listOfPitches):
	hind = -1

	nbPitches = len(listOfPitches)
	i=0
	j=i+1
	while (nbPitches>1) and (i<nbPitches-1) and ((listOfPitches[j]-listOfPitches[i])%12-7):
		if (j==nbPitches-1):
			i = i+1
			j = i+1
		else:
			j = j+1

	if (j==nbPitches-1):
		# no fifth was found, try perfect fourth
		i = 0
		j = i+1

		while (nbPitches > 1) and (i<nbPitches-1) and ((listOfPitches[j]-listOfPitches[i])%12-5):
			if (j==nbPitches-1):
				i=i+1
				j=i+1
			else:
				j=j+1

	if (i==nbPitches-1):
		hind = -1
	elif ((listOfPitches[j]-listOfPitches[i])%12==5):
		hind = listOfPitches[j]-12
	elif ((listOfPitches[j]-listOfPitches[i])%12==7):
		hind = listOfPitches[i]
	else:
		print("Hindemith problem : ", listOfPitches)
		return 0
	return hind


def virfun(listOfPitches, approxM):
	hindem = hindemith(listOfPitches)
	if hindem==-1:
		vf = virfun_gb(listOfPitches, approxM)
	else:
		vf = hindem
	return vf

def virfun_gb(listOfPitches, approxM):
	listOfFreqs = midi2freq(listOfPitches)
	approxF = midi2freqApprox(approxM)
	resu = rec_virfun(listOfFreqs, 0.1, listOfFreqs[0]*(1+approxF), approxF)
	vf = round(round(freq2midi(resu)/approxM)*approxM)
	while (vf<0):
		vf = vf+12
	return vf

def rec_virfun(freqs, divmin, divmax, approxF):
	resu = 0.0
	if (divmin<=divmax):
		if (len(freqs)==0):
			resu = (divmin+divmax)/2
		else:
			sup = freqs[0]*(1+approxF)
			inf = freqs[0]/(1+approxF)
			quo_min = ceil(inf/divmax)
			quo_max = floor(sup/divmin)
			quotient = quo_min
			while (quotient<=quo_max):
				resu = rec_virfun(freqs[2:], max(inf/quotient*1.0, divmin), min(sup/quotient*1.0, divmax), approxF)
				if (resu):
					return resu
				quotient = quotient + 1.0
	return resu

def midi2freq(midis):
	if type(midis)==type([]):
		return list(map((lambda x: 440.0*pow(2,(x-69)/12.0)), midis))
	else:
		return 440.0*pow(2,(midis-69)/12.0)

def freq2midi(freqs):
	if type(freqs)==type([]):
		return list(map((lambda x: 69 + 12*log(x/440.0, 2)), freqs))
	else:
		return 69 + 12*log(freqs/440.0, 2)

def midi2freqApprox(midiApp):
	return pow(2,midiApp/12.0)-1.0


import logging
from typing import Dict, Union, List, Optional, Tuple

from somax.classification.classifier import AbstractClassifier
from somax.runtime.activity_pattern import AbstractActivityPattern
from somax.runtime.corpus import Corpus
from somax.runtime.corpus_event import CorpusEvent
from somax.runtime.influence import AbstractInfluence, CorpusInfluence
from somax.runtime.label import AbstractLabel
from somax.runtime.memory_spaces import AbstractMemorySpace
from somax.runtime.parameter import Parametric, Parameter, ParamWithSetter
from somax.runtime.peak_event import PeakEvent
from somax.runtime.peaks import Peaks
from somax.runtime.transform_handler import TransformHandler
from somax.runtime.transforms import AbstractTransform


class Atom(Parametric):
    DEFAULT_WEIGHT = 1.0

    def __init__(self, name: str, weight: float, classifier: AbstractClassifier,
                 activity_pattern: AbstractActivityPattern, memory_space: AbstractMemorySpace, corpus: Corpus,
                 self_influenced: bool, enabled: bool = True):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self.logger.debug(f"[__init__ Creating atom '{name}'.")
        self.name = name
        self._weight: Parameter = Parameter(weight, 0.0, None, 'float', "Relative scaling of atom peaks.")
        self._enabled: Parameter = ParamWithSetter(enabled, False, True, "bool", "Enables this Atom.",
                                                   self._set_enabled)

        self._classifier: AbstractClassifier = classifier
        self._memory_space: AbstractMemorySpace = memory_space
        self._activity_pattern: AbstractActivityPattern = activity_pattern
        self._self_influenced: Parameter = Parameter(self_influenced, 0, 1, 'bool',
                                                     "Whether new events creates by player should influence this atom or not.")

        self._corpus: Optional[Corpus] = corpus
        if corpus:
            self.read_corpus(corpus)

        self._parse_parameters()

    def read_corpus(self, corpus: Optional[Corpus] = None):
        """ :raises RuntimeError """
        if corpus is not None:
            self.logger.debug(f"[read]: Reading corpus {corpus}.")
            self._corpus = corpus
        elif self._corpus is not None:
            self.logger.debug(f"[read]: Re-reading corpus {self._corpus}.")
        else:
            self.logger.debug(f"[read]: No corpus was provided and atom '{self.name}' does not have a corpus. "
                              f"No action performed.")
            return
        self._classifier.cluster(self._corpus)
        labels: List[AbstractLabel] = self._classifier.classify_corpus(self._corpus)
        self._memory_space.model(self._corpus, labels)
        self._activity_pattern.corpus = self._corpus

    # influences the memory with incoming data
    def influence(self, influence: AbstractInfluence, time: float, **kwargs) -> int:
        """ :raises InvalidLabelInput
            :returns Number of peaks generated """
        if not self.is_enabled():
            return 0

        self._update_peaks_on_influence(time)
        label: List[Tuple[AbstractLabel, AbstractTransform]] = self._classifier.classify_influence(influence)
        matched_events: List[PeakEvent] = self._memory_space.influence(label, time, **kwargs)
        if matched_events:
            self._activity_pattern.insert(matched_events)  # we insert the events into the activity profile
            return len(matched_events)
        else:
            return 0

    def _update_peaks_on_influence(self, time: float) -> None:
        self._activity_pattern.update_peaks_on_influence(time)

    def update_peaks_on_new_event(self, time: float) -> None:
        self._activity_pattern.update_peaks_on_new_event(time)

    def feedback(self, feedback_event: Optional[CorpusEvent], time: float, _applied_transform: AbstractTransform) -> None:
        if self.self_influenced and feedback_event is not None:
            self.influence(CorpusInfluence(feedback_event), time)

    def set_classifier(self, classifier: AbstractClassifier) -> None:
        self._classifier = classifier
        self.read_corpus()

    def set_memory_space(self, memory_space: AbstractMemorySpace) -> None:
        self._memory_space = memory_space
        self.read_corpus()

    def set_activity_pattern(self, activity_pattern: AbstractActivityPattern):
        activity_pattern.corpus = self._corpus
        self._activity_pattern = activity_pattern

    def update_transforms(self, transform_handler: TransformHandler):
        valid_transforms: List[AbstractTransform] = self._classifier.update_transforms(transform_handler)
        self._memory_space.update_transforms(transform_handler, valid_transforms)

    def update_parameter_dict(self) -> Dict[str, Union[Parametric, Parameter, Dict]]:
        parameters = {}
        for name, parameter in self._parse_parameters().items():
            parameters[name] = parameter.update_parameter_dict()
        self.parameter_dict = {"memory_space": self._memory_space.update_parameter_dict(),
                               "activity_pattern": self._activity_pattern.update_parameter_dict(),
                               "parameters": parameters}
        return self.parameter_dict

    @property
    def weight(self) -> float:
        return self._weight.value

    @weight.setter
    def weight(self, value: float):
        self._weight.value = value

    @property
    def classifier(self):
        return self._classifier

    @property
    def memory_space(self):
        return self._memory_space

    @property
    def self_influenced(self) -> float:
        return self._self_influenced.value

    @self_influenced.setter
    def self_influenced(self, self_influenced: bool):
        self._self_influenced.value = self_influenced

    def _set_enabled(self, enabled: bool):
        self._enabled.value = enabled
        self.clear()

    def clear(self):
        self._activity_pattern.clear()
        self._classifier.clear()
        self._memory_space.clear()

    def is_enabled(self):
        return self._enabled.value

    def get_peaks(self) -> Peaks:
        return self._activity_pattern.peaks



import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Tuple, List

from maxosc.sender import Sender, SendFormat, MaxFormatter


class SendProtocol:
    PLAYER_NUM_PEAKS = "num_peaks"
    PLAYER_SINGLE_PARAMETER = "param"
    PLAYER_CORPUS_FILES = "corpus_info"
    ALL_PLAYER_NAMES = "player_names"
    AGENT_INSTANTIATED = "instantiated_agent"
    INSTANTIATED_ATOMS = "instantiated_atoms"
    PLAYER_CORPUS = "corpus"
    PLAYER_RECORDED_CORPUS_LENGTH = "recorded_corpus"

    SCHEDULER_RUNNING = "scheduler_running"
    TRANSPORT_MODE = "transport_mode"
    SCHEDULER_CURRENT_TIME = "time"
    SCHEDULER_CURRENT_TEMPO = "tempo"
    SCHEDULER_TEMPO_SET_SUCESS = "tempo_was_set"
    SCHEDULER_HAS_TEMPO_MASTER = "has_tempo_master"
    SCHEDULER_RESET_UI = "reset"

    SEND_MIDI_EVENT = "midi"
    SEND_STATE_EVENT = "state"
    SEND_STATE_ONSET = "state_onset"

    SERVER_STATUS = "server_status"


class Target(ABC):
    WRAPPED_BANG = ["bang"]

    @abstractmethod
    def send(self, keyword: str, content: Any, **kwargs):
        raise NotImplementedError("Target.send is abstract.")


class SimpleOscTarget(Target):

    def __init__(self, address: str, port: int, ip: str = "127.0.0.1"):
        self.logger = logging.getLogger(__name__)
        self.logger.debug(f"Creating new OscTarget with address '{address}', port '{port}' and ip '{ip}'.")
        self.address: str = address
        self._client: Sender = Sender(ip, port, send_format=SendFormat.FLATTEN, warning_address="/warning")
        self._max_formatter: MaxFormatter = MaxFormatter()

    def send(self, keyword: str, content: Any, **_kwargs):
        self._client.send(self.address, [keyword, content])

    def send_dict(self, content: Dict, **_kwargs):
        max_dict: List[Tuple[str, str]] = self._max_formatter.format_maxdict_large(content)
        for address, value in max_dict:
            self.send("parameter_dict", (address, str(value)))
        self.send("parameter_dict", "bang")


class NoTarget(Target):
    def send(self, _keyword: str, _content: Any, **_kwargs):
        pass


class CallableTarget(Target):

    def __init__(self, callback_func: Callable):
        self.callback_func: Callable = callback_func

    def send(self, keyword: str, content: Any, **_kwargs):
        self.callback_func(content)



import logging
from abc import ABC
from enum import Enum
from typing import List

from somax.runtime.corpus_event import Note


class TriggerMode(Enum):
    MANUAL = "manual"
    AUTOMATIC = "automatic"

    @classmethod
    def default(cls) -> 'TriggerMode':
        return cls.MANUAL

    @classmethod
    def from_string(cls, trigger_mode: str) -> 'TriggerMode':
        try:
            return TriggerMode(trigger_mode.lower())
        except ValueError:
            logging.getLogger(__name__).warning(f"No class named '{trigger_mode} exists for the "
                                                f"{cls.__name__} module. Using default.")
            return cls.default()


class ScheduledAgent(ABC):
    def __init__(self, trigger_mode: TriggerMode, **kwargs):
        super(ScheduledAgent, self).__init__(**kwargs)
        self.trigger_mode: trigger_mode = trigger_mode


class ScheduledMidiAgent(ScheduledAgent):
    def __init__(self, trigger_mode: TriggerMode, hold_notes_artificially: bool = False,
                 simultaneous_onsets: bool = False, **kwargs):
        super().__init__(trigger_mode=trigger_mode, **kwargs)
        self.held_notes: List[Note] = []
        self.hold_notes_artificially: bool = hold_notes_artificially
        self.artificially_held_notes: List[Note] = []

        self.simultaneous_onsets: bool = simultaneous_onsets



from abc import ABC, abstractmethod


class AbstractLabel(ABC):
    @abstractmethod
    def __hash__(self):
        pass

    @abstractmethod
    def __eq__(self, other: 'AbstractLabel'):
        """Notes: Strictly not needed in current implementation but should always be implemented when __hash__ is"""
        pass


class IntLabel(AbstractLabel):
    def __init__(self, label: int):
        self.label = label

    def __repr__(self):
        return f"{self.__class__.__name__}(label={self.label})"

    def __hash__(self):
        return hash(self.label)

    def __eq__(self, other: 'AbstractLabel'):
        return isinstance(other, IntLabel) and self.label == other.label



from typing import Dict, Any

import numpy as np


from somax.settings import IMPORT_MATPLOTLIB

if IMPORT_MATPLOTLIB:
    import matplotlib.colors as colors
    import matplotlib.pyplot as plt

from somax.corpus_builder.chroma_filter import AbstractFilter, NoFilter
from somax.corpus_builder.note_matrix import NoteMatrix


class Spectrogram:
    _NUM_MIDI_ROWS = 128

    def __init__(self, spectrogram: np.ndarray, duration_ms: float, build_parameters: Dict[str, Any]):
        self.spectrogram: np.ndarray = spectrogram
        self.duration_ms: float = duration_ms
        self._build_parameters: Dict[str, Any] = build_parameters

    @classmethod
    def from_midi(cls, note_matrix: NoteMatrix, spectrogram_filter: AbstractFilter,
                  spectrogram_max_num_harmonics: int = 10, spectrogram_harmonics_decay: float = 0.5,
                  spectrogram_step_ms: float = 20.0, **_kwargs) -> 'Spectrogram':
        max_num_harmonics: int = spectrogram_max_num_harmonics
        harmonics_decay: float = spectrogram_harmonics_decay
        step_ms: float = spectrogram_step_ms

        build_parameters: Dict[str, Any] = {"max_num_harmonics": max_num_harmonics,
                                            "harmonics_decay": harmonics_decay,
                                            "step_ms": step_ms}

        note_numbers: np.ndarray = note_matrix.pitches
        num_notes: int = len(note_matrix)
        onsets_ms: np.ndarray = note_matrix.absolute_onsets
        ends_ms: np.ndarray = (note_matrix.absolute_onsets + note_matrix.absolute_durations)
        corpus_duration_ms: float = note_matrix.length_ms() + spectrogram_filter.decay_length_ms

        num_cols: int = int(np.ceil(corpus_duration_ms / step_ms)) + 1
        num_rows: int = cls._NUM_MIDI_ROWS
        spectrogram: np.ndarray = np.zeros((num_rows, num_cols))

        for i in range(num_notes):
            note_duration: int = np.ceil((ends_ms[i] - onsets_ms[i]) / step_ms).astype(int)
            if note_duration <= 0:
                continue
            note_spectrum: np.ndarray = note_numbers[i] + np.round(12 * np.log2(np.arange(1, max_num_harmonics + 1)))
            note_spectrum = note_spectrum[np.where(note_spectrum < num_rows)].astype(int)
            amplitudes: np.ndarray = np.power(harmonics_decay, np.arange(note_spectrum.size))
            envelope: np.ndarray = spectrogram_filter.filter_midi(np.ones(note_duration))
            onset_idx: int = np.round(onsets_ms[i] / step_ms)
            columns: np.ndarray = np.arange(onset_idx, onset_idx + envelope.size).astype(int)
            # TODO: Optimize if necessary: can be done with addition of sparse matrices
            # spectrogram[np.ix_(note_spectrum, columns)] += amplitudes.reshape((-1, 1)) @ envelope.reshape((1, -1))

            spectrogram[np.ix_(note_spectrum, columns)] = np.maximum(spectrogram[np.ix_(note_spectrum, columns)], amplitudes.reshape((-1, 1)) @ envelope.reshape((1, -1)))

        return cls(spectrogram, corpus_duration_ms, build_parameters)

    @classmethod
    def from_audio(cls, audio_data: np.ndarray, **kwargs):
        raise NotImplementedError("Audio Spectrogram is not implemented yet")

    # TODO: Removed for PyInstaller to exclude matplotlib
    # def plot(self, ax: Optional[Axes] = None):
    #     if not ax:
    #         fig = plt.figure()
    #         ax: Axes = fig.add_subplot(1, 1, 1)
    #
    #     ax.imshow(self.spectrogram, aspect='auto', origin='lower', norm=colors.LogNorm(vmin=0.0001, vmax=20),
    #               extent=[0, self.duration_ms, 0, 128])
    #     ax.set_ylabel("Note Number")
    #     ax.set_xlabel("Time [ms]")
    #     # plt.colorbar()

    @property
    def build_parameters(self) -> Dict[str, Any]:
        return self._build_parameters


# TODO: Implement proper Spectrogram visualization (see code below for reference)

if __name__ == '__main__':
    # import matplotlib.pyplot as plt

    filepath = '/Users/joakimborg/MIDI/debussy.mid'
    # note_matrix: NoteMatrix = NoteMatrix.from_text(np.array([60, 62, 64, 67, 71, 60]))
    n_matrix: NoteMatrix = NoteMatrix.from_midi_file(filepath)
    filt: AbstractFilter = NoFilter()

    # np.set_printoptions(linewidth=1000000, precision=3, suppress=True, threshold=1000000)
    # print(note_matrix._legacy_format()[0:10, :])
    from timeit import default_timer as timer

    # import matplotlib.colors as colors

    start = timer()
    # pd.set_option('display.max_columns', 100)
    # print(note_matrix.notes)

    spgram: Spectrogram = Spectrogram.from_midi(n_matrix, filt)
    # print(np.shape(spectrogram))
    # print(note_matrix._legacy_format())
    # spectrogram = LegacySpectrogram.computePitchClassVector(note_matrix._legacy_format())
    # print(np.shape(spectrogram))
    print(timer() - start)

    # print(spectrogram.shape)
    plt.imshow(spgram.spectrogram, aspect='auto', origin='lower', norm=colors.LogNorm(vmin=0.0001, vmax=20),
               extent=[0, spgram.duration_ms, 0, 128])
    plt.ylabel("Note Number")
    plt.xlabel("Time [ms]")
    plt.colorbar()
    plt.show()



class MemoryState:
    def __init__(self, trigger_time: float, applied_transform: 'AbstractTransform',
                 tempo: float, artificially_sustained: bool, simultaneous_onsets: bool):
        self.trigger_time: float = trigger_time
        self.applied_transform: 'AbstractTransform' = applied_transform
        self.tempo: float = tempo
        self.artificially_sustained: bool = artificially_sustained
        self.simultaneous_onsets: bool = simultaneous_onsets



import logging
from abc import abstractmethod, ABC
from typing import Dict, Union, List, Optional

import numpy as np

from somax.runtime.corpus import Corpus
from somax.runtime.parameter import Parameter, ParamWithSetter
from somax.runtime.parameter import Parametric
from somax.runtime.peak_event import PeakEvent
from somax.runtime.peaks import Peaks
from somax.utils.introspective import StringParsed


class AbstractActivityPattern(Parametric, StringParsed, ABC):
    SCORE_IDX = 0
    TIME_IDX = 1
    TRANSFORM_IDX = 2

    def __init__(self, corpus: Optional[Corpus] = None):
        super(AbstractActivityPattern, self).__init__()
        self.logger = logging.getLogger(__name__)
        self._peaks: Peaks = Peaks.create_empty()
        self.corpus: Corpus = corpus

    @classmethod
    def default(cls, **_kwargs) -> 'AbstractActivityPattern':
        return ClassicActivityPattern()

    @classmethod
    def from_string(cls, activity_pattern: str, **kwargs) -> 'AbstractActivityPattern':
        return cls._from_string(activity_pattern, **kwargs)

    @abstractmethod
    def insert(self, influences: List[PeakEvent]) -> None:
        """ """

    @abstractmethod
    def update_peaks_on_influence(self, new_time: float) -> None:
        """ """

    @abstractmethod
    def update_peaks_on_new_event(self, new_time: float) -> None:
        """ """

    @abstractmethod
    def clear(self) -> None:
        """ """

    @property
    def peaks(self) -> Peaks:
        """ Returns a shallow copy the activity pattern's peaks (copy of scores but references to times and hashes)"""
        return Peaks.optimized_copy(self._peaks)

    def update_parameter_dict(self) -> Dict[str, Union[Parametric, Parameter, Dict]]:
        parameters: Dict = {}
        for name, parameter in self._parse_parameters().items():
            parameters[name] = parameter.update_parameter_dict()
        self.parameter_dict = {"parameters": parameters}
        return self.parameter_dict


class ClassicActivityPattern(AbstractActivityPattern):
    """Decay: score = exp(-(Δt)/tau), where Δt is the time since creation in beats"""

    DEFAULT_T = 4.6

    def __init__(self, corpus: Corpus = None, tau_mem_decay: float = DEFAULT_T):
        super().__init__(corpus)
        self.logger.debug("[__init__]: ClassicActivityPattern initialized.")
        self.extinction_threshold: Parameter = Parameter(0.1, 0.0, None, 'float', "Score below which peaks are removed")
        # TODO: tau shouldn't be the parameter: t should
        self.tau_mem_decay: Parameter = ParamWithSetter(self._calc_tau(tau_mem_decay), 0, None, "float",
                                                        "Number of updates until peak is decayed below threshold.",
                                                        self._set_tau)
        self.default_score: Parameter = Parameter(1.0, None, None, 'float', "Value of a new peaks upon creation.")
        self._peaks: Peaks = Peaks.create_empty()
        self.last_update_time: float = 0.0
        self._parse_parameters()

    def insert(self, influences: List[PeakEvent]) -> None:
        self.logger.debug(f"[insert]: Inserting {len(influences)} influences.")
        scores: List[float] = []
        times: List[float] = []
        transform_hashes: List[int] = []
        for influence in influences:
            times.append(influence.event.onset)
            scores.append(self.default_score.value)
            transform_hashes.append(influence.transform_hash)
        self._peaks.append(scores, times, transform_hashes)

    def update_peaks_on_influence(self, new_time: float) -> None:
        self._update_peaks(new_time)

    def update_peaks_on_new_event(self, new_time: float) -> None:
        self._update_peaks(new_time)

    def _update_peaks(self, new_time: float) -> None:
        self._peaks.scores *= np.exp(-np.divide(new_time - self.last_update_time, self.tau_mem_decay.value))
        self._peaks.times += new_time - self.last_update_time
        self.last_update_time = new_time
        indices_to_remove: np.ndarray = np.where((self._peaks.scores <= self.extinction_threshold.value)
                                                 | (self._peaks.times >= self.corpus.duration()))
        self._peaks.remove(indices_to_remove)

    def clear(self) -> None:
        self._peaks = Peaks.create_empty()
        self.last_update_time = 0.0

    def _set_tau(self, t: float):
        self.tau_mem_decay.value = self._calc_tau(t)

    def _calc_tau(self, t: float):
        """ n is the number of updates until peak decays below threshold"""
        return -np.divide(t, np.log(self.extinction_threshold.value - 0.001))


class FreezeActivityPattern(AbstractActivityPattern):
    """Decay: score = exp(-n/tau), where n is the number of influences since creation"""

    DEFAULT_N = 3

    def __init__(self, corpus: Corpus = None):
        super().__init__(corpus)
        self.logger.debug("[__init__]: ManualActivityPattern initialized.")
        self.extinction_threshold: Parameter = Parameter(0.1, 0.0, None, 'float', "Score below which peaks are removed")
        self.tau_mem_decay: Parameter = ParamWithSetter(self._calc_tau(self.DEFAULT_N), 1, None, "int",
                                                        "Number of updates until peak is decayed below threshold.",
                                                        self._set_tau)
        self.default_score: Parameter = Parameter(1.0, None, None, 'float', "Value of a new peaks upon creation.")
        self._peaks: Peaks = Peaks.create_empty()
        self._event_indices: np.ndarray = np.zeros(0, dtype=np.int32)
        self._parse_parameters()

    def insert(self, influences: List[PeakEvent]) -> None:
        self.logger.debug(f"[insert]: Inserting {len(influences)} influences.")
        scores: List[float] = []
        times: List[float] = []
        transform_hashes: List[int] = []
        for influence in influences:
            times.append(influence.event.onset)
            scores.append(self.default_score.value)
            transform_hashes.append(influence.transform_hash)
        new_event_indices: np.ndarray = np.array([i.event.state_index for i in influences], dtype=np.int32)
        self._event_indices = np.concatenate((self._event_indices, new_event_indices))
        self._peaks.append(scores, times, transform_hashes)

    def update_peaks_on_influence(self, _new_time: float) -> None:
        if not self._peaks.is_empty():
            self._peaks.scores *= np.exp(-np.divide(1, self.tau_mem_decay.value))
            self._peaks.times += [self.corpus.event_at(i).duration for i in self._event_indices]
            self._event_indices += 1
            indices_to_remove: np.ndarray = np.where((self._peaks.scores <= self.extinction_threshold.value)
                                                     | (self._peaks.times >= self.corpus.duration())
                                                     | (self._event_indices >= self.corpus.length()))
            self._peaks.remove(indices_to_remove)
            self._event_indices = np.delete(self._event_indices, indices_to_remove)

    def update_peaks_on_new_event(self, new_time: float) -> None:
        pass

    def clear(self) -> None:
        self._peaks = Peaks.create_empty()
        self._event_indices = np.zeros(0, dtype=np.int32)

    def _set_tau(self, n: int):
        self.tau_mem_decay.value = self._calc_tau(n)

    def _calc_tau(self, n: int):
        """ n is the number of updates until peak decays below threshold"""
        return -np.divide(n, np.log(self.extinction_threshold.value - 0.001))

class ManualActivityPattern(AbstractActivityPattern):
    """Decay: score = exp(-n/tau), where n is the number of events generated since creation of peak (new_event calls)"""

    DEFAULT_N = 3
    DEFAULT_THRESHOLD_TICKS = 0.025

    def __init__(self, corpus: Corpus = None):
        super().__init__(corpus)
        self.logger.debug("[__init__]: ManualActivityPattern initialized.")
        self.extinction_threshold: Parameter = Parameter(0.1, 0.0, None, 'float', "Score below which peaks are removed")
        self.tau_mem_decay: Parameter = ParamWithSetter(self._calc_tau(self.DEFAULT_N), 1, None, "int",
                                                        "Number of updates until peak is decayed below threshold.",
                                                        self._set_tau)
        self.default_score: Parameter = Parameter(1.0, None, None, 'float', "Value of a new peaks upon creation.")
        self._peaks: Peaks = Peaks.create_empty()
        self.last_update_time: float = 0.0
        self._event_indices: np.ndarray = np.zeros(0, dtype=np.int32)
        self._parse_parameters()

    def insert(self, influences: List[PeakEvent]) -> None:
        self.logger.debug(f"[insert]: Inserting {len(influences)} influences.")
        scores: List[float] = []
        times: List[float] = []
        transform_hashes: List[int] = []
        for influence in influences:
            times.append(influence.event.onset)
            scores.append(self.default_score.value)
            transform_hashes.append(influence.transform_hash)
        new_event_indices: np.ndarray = np.array([i.event.state_index for i in influences], dtype=np.int32)
        self._event_indices = np.concatenate((self._event_indices, new_event_indices))
        self._peaks.append(scores, times, transform_hashes)

    def _update_peaks(self, new_time: float) -> None:
        if not self._peaks.is_empty() and new_time - self.last_update_time >= self.DEFAULT_THRESHOLD_TICKS:
            self._peaks.scores *= np.exp(-np.divide(1, self.tau_mem_decay.value))
            # self._peaks.times += [self.corpus.event_at(i).duration for i in self._event_indices]
            self._peaks.times = np.array([self.corpus.event_at((i + 1) % self.corpus.length()).onset
                                          for i in self._event_indices])
            self._event_indices += 1
            indices_to_remove: np.ndarray = np.where((self._peaks.scores <= self.extinction_threshold.value)
                                                     | (self._peaks.times >= self.corpus.duration())
                                                     | (self._event_indices >= self.corpus.length()))
            self._peaks.remove(indices_to_remove)
            self._event_indices = np.delete(self._event_indices, indices_to_remove)
            self.last_update_time = new_time

    def update_peaks_on_influence(self, new_time: float) -> None:
        self._update_peaks(new_time)

    def update_peaks_on_new_event(self, new_time: float) -> None:
        self._update_peaks(new_time)

    def clear(self) -> None:
        self._peaks = Peaks.create_empty()
        self._event_indices = np.zeros(0, dtype=np.int32)
        self.last_update_time = 0.0

    def _set_tau(self, n: int):
        self.tau_mem_decay.value = self._calc_tau(n)

    def _calc_tau(self, n: int):
        """ n is the number of updates until peak decays below threshold"""
        return -np.divide(n, np.log(self.extinction_threshold.value - 0.001))



from abc import ABC
from typing import Dict, Any

from somax.features import virfun
from somax.features.feature import CorpusFeature, RuntimeFeature, AbstractFeature


class AbstractIntegerPitch(AbstractFeature, ABC):
    def __init__(self, value: int):
        super().__init__(value=value)

    def value(self) -> int:
        return self._value

    @classmethod
    def decode(cls, trait_dict: Dict[str, Any]) -> 'AbstractIntegerPitch':
        return cls(value=trait_dict["pitch"])

    def encode(self) -> Dict[str, Any]:
        return {"pitch": self._value}


class RuntimeIntegerPitch(AbstractIntegerPitch, RuntimeFeature):
    @staticmethod
    def keyword() -> str:
        return "pitch"


class TopNote(AbstractIntegerPitch, CorpusFeature):
    def __init__(self, value: int):
        super().__init__(value=value)

    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        for event in corpus.events:
            event.set_feature(cls(int(max([n.pitch for n in event.notes]))))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', _fg_spectrogram: Spectrogram, _bg_spectrogram: Spectrogram,
    #             _fg_chromagram: Chromagram, _bg_chromagram: Chromagram, **_kwargs):
    #     return cls(int(max([n.pitch for n in event.notes])))


class VirtualFundamental(AbstractIntegerPitch, CorpusFeature):
    def __init__(self, value: int):
        super().__init__(value=value)

    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        for event in corpus.events:
            event.set_feature(cls(int(128 + (virfun.virfun([n.pitch for n in event.notes], 0.293) - 8) % 12)))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', _fg_spectrogram: Spectrogram, _bg_spectrogram: Spectrogram,
    #             _fg_chromagram: Chromagram, _bg_chromagram: Chromagram, **_kwargs):
    #     return cls(int(128 + (virfun.virfun([n.pitch for n in event.notes], 0.293) - 8) % 12))


class BassNote(AbstractIntegerPitch, CorpusFeature):
    def __init__(self, value: int):
        super().__init__(value=value)

    @classmethod
    def analyze(cls, corpus: 'Corpus', **kwargs) -> 'Corpus':
        for event in corpus.events:
            event.set_feature(cls(int(min([n.pitch for n in event.notes]))))
        return corpus

    # @classmethod
    # def analyze(cls, event: 'CorpusEvent', fg_spectrogram: Spectrogram, bg_spectrogram: Spectrogram,
    #             fg_chromagram: Chromagram, bg_chromagram: Chromagram, **kwargs):
    #     return cls(int(min([n.pitch for n in event.notes])))
